import nibabel as nib
import numpy as np
import logging

from dipy.io.gradients import read_bvals_bvecs
import dipy.core.gradients as dpg
from dipy.data import default_sphere

import pimms

import dipy.reconst.dki as dpy_dki
import dipy.reconst.dti as dpy_dti
import dipy.reconst.fwdti as dpy_fwdti
import dipy.reconst.msdki as dpy_msdki
from dipy.reconst.gqi import GeneralizedQSamplingModel
from dipy.reconst.rumba import RumbaSDModel, RumbaFit
from dipy.reconst import shm
from dipy.reconst.dki_micro import axonal_water_fraction

from AFQ.tasks.decorators import as_file, as_img, as_fit_deriv
from AFQ.tasks.utils import get_fname, with_name, str_to_desc
import AFQ.api.bundle_dict as abd
import AFQ.data.fetch as afd
from AFQ.utils.path import drop_extension, write_json
from AFQ._fixes import gwi_odf

from AFQ.definitions.utils import Definition
from AFQ.definitions.image import B0Image

from AFQ.models.dti import noise_from_b0
from AFQ.models.csd import _fit as csd_fit_model
from AFQ.models.csd import CsdNanResponseError
from AFQ.models.dki import _fit as dki_fit_model
from AFQ.models.dti import _fit as dti_fit_model
from AFQ.models.fwdti import _fit as fwdti_fit_model
from AFQ.models.QBallTP import (
    extract_odf, anisotropic_index, anisotropic_power)


logger = logging.getLogger('AFQ')


DIPY_GH = "https://github.com/dipy/dipy/blob/master/dipy/"


@pimms.calc("data", "gtab", "dwi", "dwi_affine")
def get_data_gtab(dwi_data_file, bval_file, bvec_file, min_bval=None,
                  max_bval=None, filter_b=True, b0_threshold=50):
    """
    DWI data as an ndarray for selected b values,
    A DIPY GradientTable with all the gradient information,
    and DWI data in a Nifti1Image,
    and the affine transformation of the DWI data.

    Parameters
    ----------
    min_bval : float, optional
        Minimum b value you want to use
        from the dataset (other than b0), inclusive.
        If None, there is no minimum limit. Default: None
    max_bval : float, optional
        Maximum b value you want to use
        from the dataset (other than b0), inclusive.
        If None, there is no maximum limit. Default: None
    filter_b : bool, optional
        Whether to filter the DWI data based on min or max bvals.
        Default: True
    b0_threshold : int, optional
        The value of b under which
        it is considered to be b0. Default: 50.
    """
    img = nib.load(dwi_data_file)
    bvals, bvecs = read_bvals_bvecs(bval_file, bvec_file)

    data = img.get_fdata()
    if filter_b and (min_bval is not None):
        valid_b = np.logical_or(
            (bvals >= min_bval), (bvals <= b0_threshold))
        data = data[..., valid_b]
        bvals = bvals[valid_b]
        bvecs = bvecs[valid_b]
    if filter_b and (max_bval is not None):
        valid_b = np.logical_or(
            (bvals <= max_bval), (bvals <= b0_threshold))
        data = data[..., valid_b]
        bvals = bvals[valid_b]
        bvecs = bvecs[valid_b]
    gtab = dpg.gradient_table(
        bvals, bvecs,
        b0_threshold=b0_threshold)
    img = nib.Nifti1Image(data, img.affine)
    return data, gtab, img, img.affine


@pimms.calc("b0")
@as_file('_desc-b0_dwi.nii.gz')
def b0(dwi, gtab):
    """
    full path to a nifti file containing the mean b0
    """
    mean_b0 = np.mean(dwi.get_fdata()[..., gtab.b0s_mask], -1)
    mean_b0_img = nib.Nifti1Image(mean_b0, dwi.affine)
    meta = dict(b0_threshold=gtab.b0_threshold)
    return mean_b0_img, meta


@pimms.calc("masked_b0")
@as_file('_desc-maskedb0_dwi.nii.gz')
def b0_mask(b0, brain_mask):
    """
    full path to a nifti file containing the
    mean b0 after applying the brain mask
    """
    img = nib.load(b0)
    brain_mask = nib.load(brain_mask).get_fdata().astype(bool)

    masked_data = img.get_fdata()
    masked_data[~brain_mask] = 0

    masked_b0_img = nib.Nifti1Image(masked_data, img.affine)
    meta = dict(
        source=b0,
        masked=True)
    return masked_b0_img, meta


@pimms.calc("dti_tf")
def dti_fit(dti_params, gtab):
    """DTI TensorFit object"""
    dti_params = nib.load(dti_params).get_fdata()
    tm = dpy_dti.TensorModel(gtab)
    return dpy_dti.TensorFit(tm, dti_params)


@pimms.calc("dti_params")
@as_file(suffix='_odfmodel-DTI_desc-diffmodel_dwi.nii.gz')
@as_img
def dti_params(brain_mask, data, gtab,
               bval_file, bvec_file, b0_threshold=50,
               robust_tensor_fitting=False):
    """
    full path to a nifti file containing parameters
    for the DTI fit

    Parameters
    ----------
    robust_tensor_fitting : bool, optional
        Whether to use robust_tensor_fitting when
        doing dti. Only applies to dti.
        Default: False
    b0_threshold : int, optional
        The value of b under which
        it is considered to be b0. Default: 50.
    """
    mask =\
        nib.load(brain_mask).get_fdata()
    if robust_tensor_fitting:
        bvals, _ = read_bvals_bvecs(
            bval_file, bvec_file)
        sigma = noise_from_b0(
            data, gtab, bvals,
            mask=mask, b0_threshold=b0_threshold)
    else:
        sigma = None
    dtf = dti_fit_model(
        gtab, data,
        mask=mask, sigma=sigma)
    meta = dict(
        Parameters=dict(
            FitMethod="WLS"),
        OutlierRejection=robust_tensor_fitting,
        ModelURL=f"{DIPY_GH}reconst/dti.py")
    return dtf.model_params, meta


@pimms.calc("fwdti_tf")
def fwdti_fit(fwdti_params, gtab):
    """Free-water DTI TensorFit object"""
    fwdti_params = nib.load(fwdti_params).get_fdata()
    fwtm = dpy_fwdti.FreeWaterTensorModel(gtab)
    return dpy_fwdti.FreeWaterTensorFit(fwtm, fwdti_params)


@pimms.calc("fwdti_params")
@as_file(suffix='_odfmodel-FWDTI_desc-diffmodel_dwi.nii.gz')
@as_img
def fwdti_params(brain_mask, data, gtab):
    """
    Full path to a nifti file containing parameters
    for the free-water DTI fit.
    """
    mask =\
        nib.load(brain_mask).get_fdata()
    dtf = fwdti_fit_model(
        data, gtab,
        mask=mask)
    meta = dict(
        Parameters=dict(
            FitMethod="NLS"),
        ModelURL=f"{DIPY_GH}reconst/fwdti.py")
    return dtf.model_params, meta


@pimms.calc("dki_tf")
def dki_fit(dki_params, gtab):
    """DKI DiffusionKurtosisFit object"""
    dki_params = nib.load(dki_params).get_fdata()
    tm = dpy_dki.DiffusionKurtosisModel(gtab)
    return dpy_dki.DiffusionKurtosisFit(tm, dki_params)


@pimms.calc("dki_params")
@as_file(suffix='_odfmodel-DKI_desc-diffmodel_dwi.nii.gz')
@as_img
def dki_params(brain_mask, gtab, data):
    """
    full path to a nifti file containing
    parameters for the DKI fit
    """
    if len(dpg.unique_bvals_magnitude(gtab.bvals)) < 3:
        raise ValueError((
            "The DKI model requires at least 2 non-zero b-values, "
            f"but you provided {len(dpg.unique_bvals_magnitude(gtab.bvals))}"
            " b-values (including b=0)."))
    mask =\
        nib.load(brain_mask).get_fdata()
    dkf = dki_fit_model(
        gtab, data,
        mask=mask)
    meta = dict(
        Parameters=dict(
            FitMethod="WLS"),
        OutlierRejection=False,
        ModelURL=f"{DIPY_GH}reconst/dki.py")
    return dkf.model_params, meta


@pimms.calc("msdki_tf")
def msdki_fit(msdki_params, gtab):
    """Mean Signal DKI DiffusionKurtosisFit object"""
    msdki_params = nib.load(msdki_params).get_fdata()
    tm = dpy_msdki.MeanDiffusionKurtosisModel(gtab)
    return dpy_msdki.MeanDiffusionKurtosisFit(tm, msdki_params)


@pimms.calc("msdki_params")
@as_file(suffix='_odfmodel-MSDKI_desc-diffmodel_dwi.nii.gz')
@as_img
def msdki_params(brain_mask, gtab, data):
    """
    full path to a nifti file containing
    parameters for the Mean Signal DKI fit
    """
    mask =\
        nib.load(brain_mask).get_fdata()
    msdki_model = dpy_msdki.MeanDiffusionKurtosisModel(gtab)
    msdki_fit = msdki_model.fit(data, mask=mask)
    meta = dict(
        ModelURL=f"{DIPY_GH}reconst/msdki.py")
    return msdki_fit.model_params, meta


@pimms.calc("msdki_msd")
@as_file('_odfmodel-MSDKI_desc-MSD_dwi.nii.gz')
@as_fit_deriv('MSDKI')
def msdki_msd(msdki_tf):
    """
    full path to a nifti file containing
    the MSDKI mean signal diffusivity
    """
    return msdki_tf.msd


@pimms.calc("msdki_msk")
@as_file('_odfmodel-MSDKI_desc-MSK_dwi.nii.gz')
@as_fit_deriv('MSDKI')
def msdki_msk(msdki_tf):
    """
    full path to a nifti file containing
    the MSDKI mean signal kurtosis
    """
    return msdki_tf.msk


@pimms.calc("csd_params")
@as_file(suffix='_odfmodel-CSD_desc-diffmodel_dwi.nii.gz')
@as_img
def csd_params(dwi, brain_mask, gtab, data,
               csd_response=None, csd_sh_order=None,
               csd_lambda_=1, csd_tau=0.1,
               csd_fa_thr=0.7):
    """
    full path to a nifti file containing
    parameters for the CSD fit

    Parameters
    ----------
    csd_response : tuple or None, optional.
        The response function to be used by CSD, as a tuple with two elements.
        The first is the eigen-values as an (3,) ndarray and the second is
        the signal value for the response function without diffusion-weighting
        (i.e. S0). If not provided, auto_response will be used to calculate
        these values.
        Default: None
    csd_sh_order : int or None, optional.
        default: infer the number of parameters from the number of data
        volumes, but no larger than 8.
        Default: None
    csd_lambda_ : float, optional.
        weight given to the constrained-positivity regularization part of
        the deconvolution equation. Default: 1
    csd_tau : float, optional.
        threshold controlling the amplitude below which the corresponding
        fODF is assumed to be zero.  Ideally, tau should be set to
        zero. However, to improve the stability of the algorithm, tau is
        set to tau*100 percent of the mean fODF amplitude (here, 10 percent
        by default)
        (see [1]_). Default: 0.1
    csd_fa_thr : float, optional.
        The threshold on the FA used to calculate the single shell auto
        response. Can be useful to reduce for baby subjects. Default: 0.7

    References
    ----------
    .. [1] Tournier, J.D., et al. NeuroImage 2007. Robust determination of
            the fibre orientation distribution in diffusion MRI:
            Non-negativity constrained super-resolved spherical
            deconvolution
    """
    mask =\
        nib.load(brain_mask).get_fdata()
    try:
        csdf = csd_fit_model(
            gtab, data,
            mask=mask,
            response=csd_response, sh_order=csd_sh_order,
            lambda_=csd_lambda_, tau=csd_tau,
            csd_fa_thr=csd_fa_thr)
    except CsdNanResponseError as e:
        raise CsdNanResponseError(
            'Could not compute CSD response function for file: '
            f'{dwi}.') from e

    meta = dict(
        SphericalHarmonicDegree=csd_sh_order,
        ResponseFunctionTensor=csd_response,
        lambda_=csd_lambda_,
        tau=csd_tau,
        csd_fa_thr=csd_fa_thr)
    meta["SphericalHarmonicBasis"] = "DESCOTEAUX"
    meta["ModelURL"] = f"{DIPY_GH}reconst/csdeconv.py"
    return csdf.shm_coeff, meta


@pimms.calc("csd_pmap")
@as_file(suffix='_odfmodel-CSD_desc-APM_dwi.nii.gz')
@as_img
def anisotropic_power_map(csd_params):
    """
    full path to a nifti file containing
    the anisotropic power map
    """
    sh_coeff = nib.load(csd_params).get_fdata()
    pmap = anisotropic_power(sh_coeff)
    return pmap, dict(CSDParamsFile=csd_params)


@pimms.calc("csd_ai")
@as_file(suffix='_odfmodel-CSD_desc-AI_dwi.nii.gz')
@as_img
def csd_anisotropic_index(csd_params):
    """
    full path to a nifti file containing
    the anisotropic index
    """
    sh_coeff = nib.load(csd_params).get_fdata()
    AI = anisotropic_index(sh_coeff)
    return AI, dict(CSDParamsFile=csd_params)


@pimms.calc("gq_params", "gq_iso", "gq_aso")
def gq(base_fname, gtab, dwi_affine, data,
       gq_sampling_length=1.2):
    """
    full path to a nifti file containing
    parameters for the Generalized Q-Sampling
    shm_coeff,
    full path to a nifti file containing isotropic diffusion component,
    full path to a nifti file containing anisotropic diffusion component

    Parameters
    ----------
    gq_sampling_length : float
        Diffusion sampling length.
        Default: 1.2
    """
    gqmodel = GeneralizedQSamplingModel(
        gtab,
        sampling_length=gq_sampling_length)

    odf = gwi_odf(gqmodel, data)

    GQ_shm, ASO, ISO = extract_odf(odf)

    params_suffix = "_odfmodel-GQ_desc-diffmodel_dwi.nii.gz"
    params_fname = get_fname(base_fname, params_suffix)
    nib.save(nib.Nifti1Image(GQ_shm, dwi_affine), params_fname)
    write_json(
        get_fname(base_fname, f"{drop_extension(params_suffix)}.json"),
        dict(GQSamplingLength=gq_sampling_length)
    )

    ASO_suffix = "_odfmodel-GQ_desc-ASO_dwi.nii.gz"
    ASO_fname = get_fname(base_fname, ASO_suffix)
    nib.save(nib.Nifti1Image(ASO, dwi_affine), ASO_fname)
    write_json(
        get_fname(base_fname, f"{drop_extension(ASO_suffix)}.json"),
        dict(GQSamplingLength=gq_sampling_length)
    )

    ISO_suffix = "_odfmodel-GQ_desc-ISO_dwi.nii.gz"
    ISO_fname = get_fname(base_fname, ISO_suffix)
    nib.save(nib.Nifti1Image(ISO, dwi_affine), ISO_fname)
    write_json(
        get_fname(base_fname, f"{drop_extension(ISO_suffix)}.json"),
        dict(GQSamplingLength=gq_sampling_length)
    )

    return params_fname, ISO_fname, ASO_fname


@pimms.calc("gq_pmap")
@as_file(suffix='_odfmodel-GQ_desc-APM_dwi.nii.gz')
@as_img
def gq_pmap(gq_params):
    """
    full path to a nifti file containing
    the anisotropic power map from GQ
    """
    sh_coeff = nib.load(gq_params).get_fdata()
    pmap = anisotropic_power(sh_coeff)
    return pmap, dict(GQParamsFile=gq_params)


@pimms.calc("gq_ai")
@as_file(suffix='_odfmodel-GQ_desc-AI_dwi.nii.gz')
@as_img
def gq_ai(gq_params):
    """
    full path to a nifti file containing
    the anisotropic index from GQ
    """
    sh_coeff = nib.load(gq_params).get_fdata()
    AI = anisotropic_index(sh_coeff)
    return AI, dict(GQParamsFile=gq_params)


@pimms.calc("rumba_model")
def rumba_model(gtab,
                rumba_wm_response=[0.0017, 0.0002, 0.0002],
                rumba_gm_response=0.0008,
                rumba_csf_response=0.003,
                rumba_n_iter=600):
    """
    fit for RUMBA-SD model as documented on dipy reconstruction options

    Parameters
    ----------
    rumba_wm_response: 1D or 2D ndarray or AxSymShResponse.
        Able to take response[0] from auto_response_ssst.
        default: array([0.0017, 0.0002, 0.0002])
    rumba_gm_response: float, optional
        Mean diffusivity for GM compartment.
        If None, then grey matter volume fraction is not computed.
        Default: 0.8e-3
    rumba_csf_response: float, optional
        Mean diffusivity for CSF compartment.
        If None, then CSF volume fraction is not computed.
        Default: 3.0e-3
    rumba_n_iter: int, optional
        Number of iterations for fODF estimation.
        Must be a positive int.
        Default: 600
    """
    return RumbaSDModel(
        gtab,
        wm_response=np.asarray(rumba_wm_response),
        gm_response=rumba_gm_response,
        csf_response=rumba_csf_response,
        n_iter=rumba_n_iter,
        recon_type='smf',
        n_coils=1,
        R=1,
        voxelwise=False,
        use_tv=False,
        sphere=default_sphere,
        verbose=True)


@pimms.calc("rumba_params")
@as_file(suffix='_odfmodel-RUMBA_desc-diffmodel_dwi.nii.gz')
@as_img
def rumba_params(rumba_model, data, brain_mask):
    """
    Takes the fitted RUMBA-SD model as input and returns
    the spherical harmonics coefficients (SHM).
    """
    rumba_fit = rumba_model.fit(
        data,
        mask=nib.load(brain_mask).get_fdata())
    odf = rumba_fit.odf(sphere=default_sphere)
    rumba_shm, _, _ = extract_odf(odf)
    meta = dict()
    return rumba_shm, meta


@pimms.calc("rumba_fit")
def rumba_fit(rumba_model, rumba_params):
    """RUMBA FIT"""
    return RumbaFit(
        rumba_model,
        nib.load(rumba_params).get_fdata())


@pimms.calc("rumba_f_csf")
@as_file(suffix='_odfmodel-RUMBA_desc-CSF_probseg.nii.gz')
@as_fit_deriv('RUMBA')
def rumba_f_csf(rumba_fit):
    """
    full path to a nifti file containing
    the CSF volume fraction for each voxel.
    """
    return rumba_fit.f_csf  # CSF volume fractions


@pimms.calc("rumba_f_gm")
@as_file(suffix='_odfmodel-RUMBA_desc-GM_probseg.nii.gz')
@as_fit_deriv('RUMBA')
def rumba_f_gm(rumba_fit):
    """
    full path to a nifti file containing
    the GM volume fraction for each voxel.
    """
    return rumba_fit.f_gm  # gray matter volume fractions


@pimms.calc("rumba_f_wm")
@as_file(suffix='_odfmodel-RUMBA_desc-WM_probseg.nii.gz')
@as_fit_deriv('RUMBA')
def rumba_f_wm(rumba_fit):
    """
    full path to a nifti file containing
    the white matter volume fraction for each voxel.
    """
    return rumba_fit.f_wm  # white matter volume fractions


@pimms.calc("opdt_params", "opdt_gfa")
def opdt_params(base_fname, data, gtab,
                dwi_affine, brain_mask,
                opdt_sh_order=8):
    """
    full path to a nifti file containing
    parameters for the Orientation Probability Density Transform
    shm_coeff,
    full path to a nifti file containing GFA

    Parameters
    ----------
    opdt_sh_order : int
        Spherical harmonics order for OPDT model. Must be even.
        Default: 8
    """
    opdt_model = shm.OpdtModel(gtab, opdt_sh_order)
    opdt_fit = opdt_model.fit(data, mask=brain_mask)

    params_suffix = "_odfmodel-OPDT_desc-diffmodel_dwi.nii.gz"
    params_fname = get_fname(base_fname, params_suffix)
    nib.save(nib.Nifti1Image(opdt_fit._shm_coef, dwi_affine), params_fname)
    write_json(
        get_fname(base_fname, f"{drop_extension(params_suffix)}.json"),
        dict(sh_order=opdt_sh_order)
    )

    GFA_suffix = "_odfmodel-OPDT_desc-GFA_dwi.nii.gz"
    GFA_fname = get_fname(base_fname, GFA_suffix)
    nib.save(nib.Nifti1Image(opdt_fit.gfa, dwi_affine), GFA_fname)
    write_json(
        get_fname(base_fname, f"{drop_extension(GFA_suffix)}.json"),
        dict(sh_order=opdt_sh_order)
    )

    return params_fname, GFA_fname


@pimms.calc("opdt_pmap")
@as_file(suffix='_odfmodel-OPDT_desc-APM_dwi.nii.gz')
@as_img
def opdt_pmap(opdt_params):
    """
    full path to a nifti file containing
    the anisotropic power map from OPDT
    """
    sh_coeff = nib.load(opdt_params).get_fdata()
    pmap = anisotropic_power(sh_coeff)
    return pmap, dict(OPDTParamsFile=opdt_params)


@pimms.calc("opdt_ai")
@as_file(suffix='_odfmodel-OPDT_desc-AI_dwi.nii.gz')
@as_img
def opdt_ai(opdt_params):
    """
    full path to a nifti file containing
    the anisotropic index from OPDT
    """
    sh_coeff = nib.load(opdt_params).get_fdata()
    AI = anisotropic_index(sh_coeff)
    return AI, dict(OPDTParamsFile=opdt_params)


@pimms.calc("csa_params", "csa_gfa")
def csa_params(base_fname, data, gtab,
               dwi_affine, brain_mask,
               csa_sh_order=8):
    """
    full path to a nifti file containing
    parameters for the Constant Solid Angle
    shm_coeff,
    full path to a nifti file containing GFA

    Parameters
    ----------
    csa_sh_order : int
        Spherical harmonics order for CSA model. Must be even.
        Default: 8
    """
    csa_model = shm.CsaOdfModel(gtab, csa_sh_order)
    csa_fit = csa_model.fit(data, mask=brain_mask)

    params_suffix = "_odfmodel-CSA_desc-diffmodel_dwi.nii.gz"
    params_fname = get_fname(base_fname, params_suffix)
    nib.save(nib.Nifti1Image(csa_fit._shm_coef, dwi_affine), params_fname)
    write_json(
        get_fname(base_fname, f"{drop_extension(params_suffix)}.json"),
        dict(sh_order=csa_sh_order)
    )

    GFA_suffix = "_odfmodel-CSA_desc-GFA_dwi.nii.gz"
    GFA_fname = get_fname(base_fname, GFA_suffix)
    nib.save(nib.Nifti1Image(csa_fit.gfa, dwi_affine), GFA_fname)
    write_json(
        get_fname(base_fname, f"{drop_extension(GFA_suffix)}.json"),
        dict(sh_order=csa_sh_order)
    )

    return params_fname, GFA_fname


@pimms.calc("csa_pmap")
@as_file(suffix='_odfmodel-CSA_desc-APM_dwi.nii.gz')
@as_img
def csa_pmap(csa_params):
    """
    full path to a nifti file containing
    the anisotropic power map from CSA
    """
    sh_coeff = nib.load(csa_params).get_fdata()
    pmap = anisotropic_power(sh_coeff)
    return pmap, dict(CSAParamsFile=csa_params)


@pimms.calc("csa_ai")
@as_file(suffix='_odfmodel-CSA_desc-AI_dwi.nii.gz')
@as_img
def csa_ai(csa_params):
    """
    full path to a nifti file containing
    the anisotropic index from CSA
    """
    sh_coeff = nib.load(csa_params).get_fdata()
    AI = anisotropic_index(sh_coeff)
    return AI, dict(CSAParamsFile=csa_params)


@pimms.calc("fwdti_fa")
@as_file(suffix='_odfmodel-FWDTI_desc-FA_dwi.nii.gz')
@as_fit_deriv('FWDTI')
def fwdti_fa(fwdti_tf):
    """
    full path to a nifti file containing the Free-water DTI fractional
    anisotropy
    """
    return fwdti_tf.fa


@pimms.calc("fwdti_md")
@as_file(suffix='_odfmodel-FWDTI_desc-MD_dwi.nii.gz')
@as_fit_deriv('FWDTI')
def fwdti_md(fwdti_tf):
    """
    full path to a nifti file containing the Free-water DTI mean diffusivity
    """
    return fwdti_tf.md


@pimms.calc("fwdti_fwf")
@as_file(suffix='_odfmodel-FWDTI_desc-FWF_dwi.nii.gz')
@as_fit_deriv('FWDTI')
def fwdti_fwf(fwdti_tf):
    """
    full path to a nifti file containing the Free-water DTI free water fraction
    """
    return fwdti_tf.f


@pimms.calc("dti_fa")
@as_file(suffix='_odfmodel-DTI_desc-FA_dwi.nii.gz')
@as_fit_deriv('DTI')
def dti_fa(dti_tf):
    """
    full path to a nifti file containing
    the DTI fractional anisotropy
    """
    return dti_tf.fa


@pimms.calc("dti_lt0", "dti_lt1", "dti_lt2", "dti_lt3", "dti_lt4", "dti_lt5")
def dti_lt(dti_tf, dwi_affine):
    """
    Image of first element in the DTI tensor according to DIPY convention
    i.e. Dxx (rate of diffusion from the left to right side of the brain),
    Image of second element in the DTI tensor according to DIPY convention
    i.e. Dyy (rate of diffusion from the posterior to anterior part of 
    the brain),
    Image of third element in the DTI tensor according to DIPY convention
    i.e. Dzz (rate of diffusion from the inferior to superior part of the
    brain),
    Image of fourth element in the DTI tensor according to DIPY convention
    i.e. Dxy (rate of diffusion in the xy plane indicating the 
    relationship between the x and y directions),
    Image of fifth element in the DTI tensor according to DIPY convention
    i.e. Dxz (rate of diffusion in the xz plane indicating the
    relationship between the x and z directions),
    Image of sixth element in the DTI tensor according to DIPY convention
    i.e. Dyz (rate of diffusion in the yz plane indicating the
    relationship between the y and z directions)
    """
    dti_lt_dict = {}
    for ii in range(6):
        dti_lt_dict[f"dti_lt{ii}"] = nib.Nifti1Image(
            dti_tf.lower_triangular()[..., ii],
            dwi_affine)
    return dti_lt_dict


@pimms.calc("dti_cfa")
@as_file(suffix='_odfmodel-DTI_desc-CFA_dwi.nii.gz')
@as_fit_deriv('DTI')
def dti_cfa(dti_tf):
    """
    full path to a nifti file containing
    the DTI color fractional anisotropy
    """
    return dti_tf.color_fa


@pimms.calc("dti_pdd")
@as_file(suffix='_odfmodel-DTI_desc-PDD_dwi.nii.gz')
@as_fit_deriv('DTI')
def dti_pdd(dti_tf):
    """
    full path to a nifti file containing
    the DTI principal diffusion direction
    """
    pdd = dti_tf.directions.squeeze()
    # Invert the x coordinates:
    pdd[..., 0] = pdd[..., 0] * -1
    return pdd


@pimms.calc("dti_md")
@as_file('_odfmodel-DTI_desc-MD_dwi.nii.gz')
@as_fit_deriv('DTI')
def dti_md(dti_tf):
    """
    full path to a nifti file containing
    the DTI mean diffusivity
    """
    return dti_tf.md


@pimms.calc("dti_ga")
@as_file(suffix='_odfmodel-DTI_desc-GA_dwi.nii.gz')
@as_fit_deriv('DTI')
def dti_ga(dti_tf):
    """
    full path to a nifti file containing
    the DTI geodesic anisotropy
    """
    return dti_tf.ga


@pimms.calc("dti_rd")
@as_file(suffix='_odfmodel-DTI_desc-RD_dwi.nii.gz')
@as_fit_deriv('DTI')
def dti_rd(dti_tf):
    """
    full path to a nifti file containing
    the DTI radial diffusivity
    """
    return dti_tf.rd


@pimms.calc("dti_ad")
@as_file(suffix='_odfmodel-DTI_desc-AD_dwi.nii.gz')
@as_fit_deriv('DTI')
def dti_ad(dti_tf):
    """
    full path to a nifti file containing
    the DTI axial diffusivity
    """
    return dti_tf.ad


@pimms.calc(
    "dki_kt0", "dki_kt1", "dki_kt2", "dki_kt3", "dki_kt4",
    "dki_kt5", "dki_kt6", "dki_kt7", "dki_kt8", "dki_kt9",
    "dki_kt10", "dki_kt11", "dki_kt12", "dki_kt13", "dki_kt14")
def dki_kt(dki_tf, dwi_affine):
    """
    Image of first element in the DKI kurtosis model,
    Image of second element in the DKI kurtosis model,
    Image of third element in the DKI kurtosis model,
    Image of fourth element in the DKI kurtosis model,
    Image of fifth element in the DKI kurtosis model,
    Image of sixth element in the DKI kurtosis model,
    Image of seventh element in the DKI kurtosis model,
    Image of eighth element in the DKI kurtosis model,
    Image of ninth element in the DKI kurtosis model,
    Image of tenth element in the DKI kurtosis model,
    Image of eleventh element in the DKI kurtosis model,
    Image of twelfth element in the DKI kurtosis model,
    Image of thirteenth element in the DKI kurtosis model,
    Image of fourteenth element in the DKI kurtosis model,
    Image of fifteenth element in the DKI kurtosis model
    """
    dki_kt_dict = {}
    for ii in range(15):
        dki_kt_dict[f"dki_kt{ii}"] = nib.Nifti1Image(
            dki_tf.kt[..., ii],
            dwi_affine)
    return dki_kt_dict


@pimms.calc("dki_lt0", "dki_lt1", "dki_lt2", "dki_lt3", "dki_lt4", "dki_lt5")
def dki_lt(dki_tf, dwi_affine):
    """
    Image of first element in the DTI tensor from DKI,
    Image of second element in the DTI tensor from DKI,
    Image of third element in the DTI tensor from DKI,
    Image of fourth element in the DTI tensor from DKI,
    Image of fifth element in the DTI tensor from DKI,
    Image of sixth element in the DTI tensor from DKI
    """
    dki_lt_dict = {}
    for ii in range(6):
        dki_lt_dict[f"dki_lt{ii}"] = nib.Nifti1Image(
            dki_tf.lower_triangular()[..., ii],
            dwi_affine)
    return dki_lt_dict


@pimms.calc("dki_fa")
@as_file('_odfmodel-DKI_desc-FA_dwi.nii.gz')
@as_fit_deriv('DKI')
def dki_fa(dki_tf):
    """
    full path to a nifti file containing
    the DKI fractional anisotropy
    """
    return dki_tf.fa


@pimms.calc("dki_md")
@as_file('_odfmodel-DKI_desc-MD_dwi.nii.gz')
@as_fit_deriv('DKI')
def dki_md(dki_tf):
    """
    full path to a nifti file containing
    the DKI mean diffusivity
    """
    return dki_tf.md


@pimms.calc("dki_awf")
@as_file('_odfmodel-DKI_desc-AWF_dwi.nii.gz')
@as_fit_deriv('DKI')
def dki_awf(dki_params,
            sphere='repulsion100', gtol=1e-2):
    """
    full path to a nifti file containing
    the DKI axonal water fraction

    Parameters
    ----------
    sphere : Sphere class instance, optional
        The sphere providing sample directions for the initial
        search of the maximal value of kurtosis.
        Default: 'repulsion100'
    gtol : float, optional
        This input is to refine kurtosis maxima under the precision of
        the directions sampled on the sphere class instance.
        The gradient of the convergence procedure must be less than gtol
        before successful termination.
        If gtol is None, fiber direction is directly taken from the initial
        sampled directions of the given sphere object.
        Default: 1e-2
    """
    dki_params = nib.load(dki_params).get_fdata()
    return axonal_water_fraction(dki_params, sphere=sphere, gtol=gtol)


@pimms.calc("dki_mk")
@as_file('_odfmodel-DKI_desc-MK_dwi.nii.gz')
@as_fit_deriv('DKI')
def dki_mk(dki_tf):
    """
    full path to a nifti file containing
    the DKI mean kurtosis file
    """
    return dki_tf.mk()


@pimms.calc("dki_kfa")
@as_file('_odfmodel-DKI_desc-KFA_dwi.nii.gz')
@as_fit_deriv('DKI')
def dki_kfa(dki_tf):
    """
    full path to a nifti file containing
    the DKI kurtosis FA file

    References
    ----------
    .. [Hansen2019] Hansen B. An Introduction to Kurtosis Fractional
    Anisotropy. AJNR Am J Neuroradiol. 2019 Oct;40(10):1638-1641.
    doi: 10.3174/ajnr.A6235. Epub 2019 Sep 26. PMID: 31558496;
    PMCID: PMC7028548.
    """
    return dki_tf.kfa


@pimms.calc("dki_ga")
@as_file(suffix='_odfmodel-DKI_desc-GA_dwi.nii.gz')
@as_fit_deriv('DKI')
def dki_ga(dki_tf):
    """
    full path to a nifti file containing
    the DKI geodesic anisotropy
    """
    return dki_tf.ga


@pimms.calc("dki_rd")
@as_file(suffix='_odfmodel-DKI_desc-RD_dwi.nii.gz')
@as_fit_deriv('DKI')
def dki_rd(dki_tf):
    """
    full path to a nifti file containing
    the DKI radial diffusivity
    """
    return dki_tf.rd


@pimms.calc("dki_ad")
@as_file(suffix='_odfmodel-DKI_desc-AD_dwi.nii.gz')
@as_fit_deriv('DKI')
def dki_ad(dki_tf):
    """
    full path to a nifti file containing
    the DKI axial diffusivity
    """
    return dki_tf.ad


@pimms.calc("dki_rk")
@as_file(suffix='_odfmodel-DKI_desc-RK_dwi.nii.gz')
@as_fit_deriv('DKI')
def dki_rk(dki_tf):
    """
    full path to a nifti file containing
    the DKI radial kurtosis
    """
    return dki_tf.rk


@pimms.calc("dki_ak")
@as_file(suffix='_odfmodel-DKI_desc-AK_dwi.nii.gz')
@as_fit_deriv('DKI')
def dki_ak(dki_tf):
    """
    full path to a nifti file containing
    the DKI axial kurtosis file
    """
    return dki_tf.ak


@pimms.calc("brain_mask")
@as_file('_desc-brain_mask.nii.gz')
def brain_mask(b0, brain_mask_definition=None):
    """
    full path to a nifti file containing
    the brain mask

    Parameters
    ----------
    brain_mask_definition : instance from `AFQ.definitions.image`, optional
        This will be used to create
        the brain mask, which gets applied before registration to a
        template.
        If you want no brain mask to be applied, use FullImage.
        If None, use B0Image()
        Default: None
    """
    # Note that any case where brain_mask_definition is not None
    # is handled in get_data_plan
    # This is just the default
    return B0Image().get_image_getter("data")(b0)


@pimms.calc("bundle_dict", "reg_template")
def get_bundle_dict(segmentation_params,
                    brain_mask, b0,
                    bundle_info=None, reg_template_spec="mni_T1"):
    """
    Dictionary defining the different bundles to be segmented,
    and a Nifti1Image containing the template for registration

    Parameters
    ----------
    bundle_info : dict or BundleDict, optional
        A dictionary or BundleDict for use in segmentation.
        See `Defining Custom Bundle Dictionaries`
        in the `usage` section of pyAFQ's documentation for details.
        If None, will get all appropriate bundles for the chosen
        segmentation algorithm.
        Default: None
    reg_template_spec : str, or Nifti1Image, optional
        The target image data for registration.
        Can either be a Nifti1Image, a path to a Nifti1Image, or
        if "mni_T2", "dti_fa_template", "hcp_atlas", or "mni_T1",
        image data will be loaded automatically.
        If "hcp_atlas" is used, slr registration will be used
        and reg_subject should be "subject_sls".
        Default: "mni_T1"
    """
    if not isinstance(reg_template_spec, str)\
            and not isinstance(reg_template_spec, nib.Nifti1Image):
        raise TypeError(
            "reg_template must be a str or Nifti1Image")

    if bundle_info is not None and not ((
            isinstance(bundle_info, dict)) or (
            isinstance(bundle_info, abd.BundleDict))):
        raise TypeError((
            "bundle_info must be"
            " a dict, or a BundleDict"))

    if bundle_info is None:
        bundle_info = abd.default18_bd() + abd.callosal_bd()

    use_brain_mask = True
    brain_mask = nib.load(brain_mask).get_fdata()
    if np.all(brain_mask == 1.0):
        use_brain_mask = False
    if isinstance(reg_template_spec, nib.Nifti1Image):
        reg_template = reg_template_spec
    else:
        img_l = reg_template_spec.lower()
        if img_l == "mni_t2":
            reg_template = afd.read_mni_template(
                mask=use_brain_mask, weight="T2w")
        elif img_l == "mni_t1":
            reg_template = afd.read_mni_template(
                mask=use_brain_mask, weight="T1w")
        elif img_l == "dti_fa_template":
            reg_template = afd.read_ukbb_fa_template(mask=use_brain_mask)
        elif img_l == "hcp_atlas":
            reg_template = afd.read_mni_template(mask=use_brain_mask)
        elif img_l == "pediatric":
            reg_template = afd.read_pediatric_templates()[
                "UNCNeo-withCerebellum-for-babyAFQ"]
        else:
            reg_template = nib.load(reg_template_spec)

    if isinstance(bundle_info, abd.BundleDict):
        bundle_dict = bundle_info.copy()
    else:
        bundle_dict = abd.BundleDict(
            bundle_info,
            resample_to=reg_template)

    return bundle_dict, reg_template


def get_data_plan(kwargs):
    if "scalars" in kwargs and not (
        isinstance(kwargs["scalars"], list) and isinstance(
            kwargs["scalars"][0], (str, Definition))):
        raise TypeError(
            "scalars must be a list of "
            "strings/scalar definitions")

    data_tasks = with_name([
        get_data_gtab, b0, b0_mask, brain_mask,
        dti_fit, dki_fit, fwdti_fit, anisotropic_power_map,
        csd_anisotropic_index,
        dti_fa, dti_lt, dti_cfa, dti_pdd, dti_md, dki_kt, dki_lt, dki_fa,
        gq, gq_pmap, gq_ai, opdt_params, opdt_pmap, opdt_ai,
        csa_params, csa_pmap, csa_ai,
        fwdti_fa, fwdti_md, fwdti_fwf,
        msdki_fit, msdki_params, msdki_msd, msdki_msk,
        dki_md, dki_awf, dki_mk, dki_kfa, dki_ga, dki_rd,
        dti_ga, dti_rd, dti_ad,
        dki_ad, dki_rk, dki_ak, dti_params, dki_params, fwdti_params,
        rumba_fit, rumba_params, rumba_model,
        rumba_f_csf, rumba_f_gm, rumba_f_wm,
        csd_params, get_bundle_dict])

    if "scalars" not in kwargs:
        bvals, _ = read_bvals_bvecs(kwargs["bval_file"], kwargs["bvec_file"])
        if len(dpg.unique_bvals_magnitude(bvals)) > 2:
            kwargs["scalars"] = [
                "dki_fa", "dki_md",
                "dki_kfa", "dki_mk"]
        else:
            kwargs["scalars"] = [
                "dti_fa", "dti_md"]
    else:
        scalars = []
        for scalar in kwargs["scalars"]:
            if isinstance(scalar, str):
                scalars.append(scalar.lower())
            else:
                scalars.append(scalar)
        kwargs["scalars"] = scalars

    bm_def = kwargs.get(
        "brain_mask_definition", None)
    if bm_def is not None:
        if not isinstance(bm_def, Definition):
            raise TypeError(
                "brain_mask_definition must be a Definition")
        data_tasks["brain_mask_res"] = pimms.calc("brain_mask")(
            as_file((
                f'_desc-{str_to_desc(bm_def.get_name())}'
                '_dwi.nii.gz'))(bm_def.get_image_getter("data")))

    return pimms.plan(**data_tasks)



import nibabel as nib
import logging
import numpy as np
import os
import os.path as op
from time import time
import pandas as pd

import pimms

from dipy.align import resample

from AFQ.tasks.utils import get_fname, with_name, str_to_desc
import AFQ.utils.volume as auv
from AFQ.viz.utils import Viz
import AFQ.utils.streamlines as aus
from AFQ.utils.path import write_json

from plotly.subplots import make_subplots

logger = logging.getLogger('AFQ')


def _viz_prepare_vol(vol, xform, mapping, scalar_dict):
    if vol in scalar_dict.keys():
        vol = scalar_dict[vol]
        if isinstance(vol, str):
            vol = nib.load(vol)
        vol = vol.get_fdata()
    if isinstance(vol, str):
        vol = nib.load(vol).get_fdata()
    if xform:
        vol = mapping.transform_inverse(vol)
    vol[np.isnan(vol)] = 0
    return vol


@pimms.calc("all_bundles_figure")
def viz_bundles(base_fname,
                viz_backend,
                data_imap,
                mapping_imap,
                segmentation_imap,
                tracking_params,
                segmentation_params,
                best_scalar,
                sbv_lims_bundles=[None, None],
                volume_opacity_bundles=0.3,
                n_points_bundles=40):
    """
    figure for the visualizaion of the recognized
    bundles in the subject's brain.

    Parameters
    ----------
    sbv_lims_bundles : ndarray
        Of the form (lower bound, upper bound). Shading based on
        shade_by_volume will only differentiate values within these bounds.
        If lower bound is None, will default to 0.
        If upper bound is None, will default to the maximum value in
        shade_by_volume.
        Default: [None, None]
    volume_opacity_bundles : float, optional
        Opacity of volume slices.
        Default: 0.3
    n_points_bundles : int or None
        n_points to resample streamlines to before plotting. If None, no
        resampling is done.
        Default: 40

    Returns
    -------
    List of Figure, String or just the Figure:
    If file can be generated, returns a tuple including the figure and the
    path to the file.
    Otherwise, returns the figure.
    """
    mapping = mapping_imap["mapping"]
    scalar_dict = segmentation_imap["scalar_dict"]
    profiles_file = segmentation_imap["profiles"]
    volume = data_imap["masked_b0"]
    shade_by_volume = data_imap[best_scalar]
    start_time = time()
    volume = _viz_prepare_vol(volume, False, mapping, scalar_dict)
    shade_by_volume = _viz_prepare_vol(
        shade_by_volume, False, mapping, scalar_dict)

    flip_axes = [False, False, False]
    for i in range(3):
        flip_axes[i] = (data_imap["dwi_affine"][i, i] < 0)

    if "plotly" in viz_backend.backend:
        figure = make_subplots(
            rows=1, cols=2,
            specs=[[{"type": "scene"}, {"type": "scene"}]])
    else:
        figure = None

    figure = viz_backend.visualize_volume(
        volume,
        opacity=volume_opacity_bundles,
        flip_axes=flip_axes,
        interact=False,
        inline=False,
        figure=figure)

    figure = viz_backend.visualize_bundles(
        segmentation_imap["bundles"],
        shade_by_volume=shade_by_volume,
        sbv_lims=sbv_lims_bundles,
        include_profiles=(pd.read_csv(profiles_file), best_scalar),
        n_points=n_points_bundles,
        flip_axes=flip_axes,
        interact=False,
        inline=False,
        figure=figure)

    fname = None
    if "no_gif" not in viz_backend.backend:
        fname = get_fname(
            base_fname, '_desc-viz_dwi.gif',
            tracking_params=tracking_params,
            segmentation_params=segmentation_params)

        viz_backend.create_gif(figure, fname)
    if "plotly" in viz_backend.backend:
        fname = get_fname(
            base_fname, '_desc-viz_dwi.html',
            tracking_params=tracking_params,
            segmentation_params=segmentation_params)

        figure.write_html(fname)
    meta_fname = get_fname(
        base_fname, '_desc-viz_dwi.json',
        tracking_params=tracking_params,
        segmentation_params=segmentation_params)
    meta = dict(Timing=time() - start_time)
    write_json(meta_fname, meta)
    if fname is None:
        return figure
    else:
        return [figure, fname]


@pimms.calc("indiv_bundles_figures")
def viz_indivBundle(base_fname,
                    output_dir,
                    viz_backend,
                    data_imap,
                    mapping_imap,
                    segmentation_imap,
                    tracking_params,
                    segmentation_params,
                    best_scalar,
                    sbv_lims_indiv=[None, None],
                    volume_opacity_indiv=0.3,
                    n_points_indiv=40):
    """
    list of full paths to html or gif files
    containing visualizaions of individual bundles

    Parameters
    ----------
    sbv_lims_indiv : ndarray
        Of the form (lower bound, upper bound). Shading based on
        shade_by_volume will only differentiate values within these bounds.
        If lower bound is None, will default to 0.
        If upper bound is None, will default to the maximum value in
        shade_by_volume.
        Default: [None, None]
    volume_opacity_indiv : float, optional
        Opacity of volume slices.
        Default: 0.3
    n_points_indiv : int or None
        n_points to resample streamlines to before plotting. If None, no
        resampling is done.
        Default: 40
    """
    mapping = mapping_imap["mapping"]
    bundle_dict = data_imap["bundle_dict"]
    reg_template = data_imap["reg_template"]
    scalar_dict = segmentation_imap["scalar_dict"]
    volume = data_imap["masked_b0"]
    shade_by_volume = data_imap[best_scalar]
    profiles = pd.read_csv(segmentation_imap["profiles"])

    start_time = time()
    volume = _viz_prepare_vol(
        volume, False, mapping, scalar_dict)
    shade_by_volume = _viz_prepare_vol(
        shade_by_volume, False, mapping, scalar_dict)

    flip_axes = [False, False, False]
    for i in range(3):
        flip_axes[i] = (data_imap["dwi_affine"][i, i] < 0)

    bundles = aus.SegmentedSFT.fromfile(
        segmentation_imap["bundles"])

    # This dictionary contains a mapping to which ROIs
    # should be used from the bundle dict, based on the
    # name from the segmented SFT file. Currently,
    # This is only different when using bundle sections.
    segmented_bname_to_roi_bname = {}
    for b_name, b_info in bundle_dict.items():
        if "bundlesection" in b_info:
            for sb_name in b_info["bundlesection"]:
                segmented_bname_to_roi_bname[sb_name] = b_name
        else:
            segmented_bname_to_roi_bname[b_name] = b_name

    figures = {}
    for bundle_name in bundles.bundle_names:
        logger.info(f"Generating {bundle_name} visualization...")
        roi_bname = segmented_bname_to_roi_bname[bundle_name]

        figure = viz_backend.visualize_volume(
            volume,
            opacity=volume_opacity_indiv,
            flip_axes=flip_axes,
            interact=False,
            inline=False)
        if len(bundles.get_bundle(bundle_name)) > 0:
            figure = viz_backend.visualize_bundles(
                bundles,
                shade_by_volume=shade_by_volume,
                sbv_lims=sbv_lims_indiv,
                bundle=bundle_name,
                n_points=n_points_indiv,
                flip_axes=flip_axes,
                interact=False,
                inline=False,
                figure=figure)
        else:
            logger.info(
                "No streamlines found to visualize for "
                + bundle_name)

        warped_rois = []
        for reg_type in ['start', 'end']:
            if reg_type in bundle_dict[
                    roi_bname]:
                pp = bundle_dict[roi_bname][reg_type]
                pp = resample(
                    pp.get_fdata(),
                    reg_template,
                    pp.affine,
                    reg_template.affine).get_fdata()

                atlas_roi = np.zeros(pp.shape)
                atlas_roi[np.where(pp > 0)] = 1
                warped_roi = auv.transform_inverse_roi(
                    atlas_roi,
                    mapping,
                    bundle_name=roi_bname)
                warped_rois.append(warped_roi)
        for i, roi in enumerate(warped_rois):
            figure = viz_backend.visualize_roi(
                roi,
                name=f"{roi_bname} endpoint ROI {i}",
                flip_axes=flip_axes,
                inline=False,
                interact=False,
                figure=figure)

        for roi_fname in mapping_imap["rois"][roi_bname]:
            figure = viz_backend.visualize_roi(
                roi_fname,
                name=roi_fname.split("desc-")[1].split("_")[0],
                flip_axes=flip_axes,
                inline=False,
                interact=False,
                figure=figure)

        roi_dir = op.join(output_dir, 'viz_bundles')
        os.makedirs(roi_dir, exist_ok=True)
        figures[bundle_name] = figure
        if "no_gif" not in viz_backend.backend:
            fname = op.split(
                get_fname(
                    base_fname,
                    f'_desc-{str_to_desc(bundle_name)}viz'
                    f'_dwi.gif',
                    tracking_params=tracking_params,
                    segmentation_params=segmentation_params))

            fname = op.join(roi_dir, fname[1])
            viz_backend.create_gif(figure, fname)
        if "plotly" in viz_backend.backend:
            roi_dir = op.join(output_dir, 'viz_bundles')
            os.makedirs(roi_dir, exist_ok=True)
            fname = op.split(
                get_fname(
                    base_fname,
                    f'_desc-{str_to_desc(bundle_name)}viz'
                    f'_dwi.html',
                    tracking_params=tracking_params,
                    segmentation_params=segmentation_params))

            fname = op.join(roi_dir, fname[1])
            figure.write_html(fname)

            # also do the core visualizations when using the plotly backend
            core_dir = op.join(output_dir, 'viz_core_bundles')
            os.makedirs(core_dir, exist_ok=True)
            indiv_profile = profiles[
                profiles.tractID == bundle_name][best_scalar].to_numpy()
            if len(indiv_profile) > 1:
                fname = op.split(
                    get_fname(
                        base_fname,
                        f'_desc-{str_to_desc(bundle_name)}viz'
                        f'_dwi.html',
                        tracking_params=tracking_params,
                        segmentation_params=segmentation_params))
                fname = op.join(core_dir, fname[1])
                core_fig = make_subplots(
                    rows=1, cols=2,
                    specs=[[{"type": "scene"}, {"type": "scene"}]])
                core_fig = viz_backend.visualize_volume(
                    volume,
                    opacity=volume_opacity_indiv,
                    flip_axes=flip_axes,
                    figure=core_fig,
                    interact=False,
                    inline=False)
                core_fig = viz_backend.visualize_bundles(
                    segmentation_imap["bundles"],
                    shade_by_volume=shade_by_volume,
                    sbv_lims=sbv_lims_indiv,
                    bundle=bundle_name,
                    colors={bundle_name: [0.5, 0.5, 0.5]},
                    n_points=n_points_indiv,
                    flip_axes=flip_axes,
                    interact=False,
                    inline=False,
                    figure=core_fig)
                core_fig = viz_backend.single_bundle_viz(
                    indiv_profile,
                    segmentation_imap["bundles"],
                    bundle_name,
                    best_scalar,
                    flip_axes=flip_axes,
                    figure=core_fig,
                    include_profile=True)
                core_fig.write_html(fname)
    meta_fname = get_fname(
        base_fname, f'_desc-{str_to_desc(bundle_name)}viz_dwi',
        tracking_params=tracking_params,
        segmentation_params=segmentation_params)
    meta = dict(Timing=time() - start_time)
    write_json(meta_fname, meta)
    return {"indiv_bundles_figures": figures}


@pimms.calc("tract_profile_plots")
def plot_tract_profiles(base_fname, scalars, tracking_params,
                        segmentation_params, segmentation_imap):
    """
    list of full paths to png files,
    where files contain plots of the tract profiles
    """
    from AFQ.viz.plot import visualize_tract_profiles
    start_time = time()
    fnames = []
    for scalar in scalars:
        this_scalar = scalar if isinstance(scalar, str) else scalar.get_name()
        fname = get_fname(
            base_fname,
            f'_model-{str_to_desc(this_scalar)}_desc-vizprofile_dwi',
            tracking_params=tracking_params,
            segmentation_params=segmentation_params)
        tract_profiles_folder = op.join(
            op.dirname(fname),
            "tract_profile_plots")
        fname = op.join(
            tract_profiles_folder,
            op.basename(fname))
        os.makedirs(op.abspath(tract_profiles_folder), exist_ok=True)

        visualize_tract_profiles(
            segmentation_imap["profiles"],
            scalar=this_scalar,
            file_name=fname,
            n_boot=100)
        fnames.append(fname + ".png")
        meta_fname = fname + ".json"
        meta = dict(Timing=time() - start_time)
        write_json(meta_fname, meta)
    return fnames


@pimms.calc("viz_backend")
def init_viz_backend(viz_backend_spec="plotly_no_gif",
                     virtual_frame_buffer=False):
    """
    An instance of the `AFQ.viz.utils.viz_backend` class.

    Parameters
    ----------
    virtual_frame_buffer : bool, optional
        Whether to use a virtual fram buffer. This is neccessary if
        generating GIFs in a headless environment. Default: False
    viz_backend_spec : str, optional
        Which visualization backend to use.
        See Visualization Backends page in documentation for details:
        https://yeatmanlab.github.io/pyAFQ/usage/viz_backend.html
        One of {"fury", "plotly", "plotly_no_gif"}.
        Default: "plotly_no_gif"
    """
    if not isinstance(virtual_frame_buffer, bool):
        raise TypeError("virtual_frame_buffer must be a bool")
    if "fury" not in viz_backend_spec\
            and "plotly" not in viz_backend_spec:
        raise TypeError(
            "viz_backend_spec must contain either 'fury' or 'plotly'")

    if virtual_frame_buffer:
        from xvfbwrapper import Xvfb
        vdisplay = Xvfb(width=1280, height=1280)
        vdisplay.start()

    return Viz(backend=viz_backend_spec.lower())


def get_viz_plan(kwargs):
    viz_tasks = with_name([
        plot_tract_profiles, viz_bundles, viz_indivBundle, init_viz_backend])
    return pimms.plan(**viz_tasks)



import os.path as op
import os
import logging
import tempfile

import numpy as np
from scipy.stats import sem
import pandas as pd
from tqdm.auto import tqdm

import AFQ.viz.utils as vut
from AFQ.viz.utils import display_string
from AFQ.utils.stats import contrast_index as calc_contrast_index
from AFQ.data.utils import BUNDLE_RECO_2_AFQ, BUNDLE_MAT_2_PYTHON

try:
    from pingouin import intraclass_corr, corr
    import seaborn as sns
    import matplotlib.pyplot as plt
    from matplotlib.lines import Line2D
    from matplotlib.patches import Patch
    import IPython.display as display
except (ImportError, ModuleNotFoundError):
    raise ImportError(vut.viz_import_msg_error("plot"))

__all__ = ["visualize_tract_profiles", "visualize_gif_inline"]


def visualize_tract_profiles(tract_profiles, scalar="dti_fa", ylim=None,
                             n_boot=1000,
                             file_name=None,
                             positions=vut.POSITIONS):
    """
    Visualize all tract profiles for a scalar in one plot

    Parameters
    ----------
    tract_profiles : string
        Path to CSV containing tract_profiles.

    scalar : string, optional
       Scalar to use in plots. Default: "dti_fa".

    ylim : list of 2 floats, optional
        Minimum and maximum value used for y-axis bounds.
        If None, ylim is not set.
        Default: None

    n_boot : int, optional
        Number of bootstrap resamples for seaborn to use
        to estimate the ci.
        Default: 1000

    file_name : string, optional
        File name to save figure to if not None. Default: None

    positions : dictionary, optional
        Dictionary that maps bundle names to position in plot.
        Default: vut.POSITIONS

    Returns
    -------
        Matplotlib figure and axes.
    """
    csv_comparison = GroupCSVComparison(
        None,
        [tract_profiles],
        ["my_tract_profiles"],
        remove_model=False,
        scalar_bounds={'lb': {}, 'ub': {}})

    df = csv_comparison.tract_profiles(
        scalar=scalar,
        ylim=ylim,
        positions=positions,
        out_file=file_name,
        plot_subject_lines=False,
        n_boot=n_boot)

    return df


class BrainAxes():
    '''
    Helper class.
    Creates and handles a grid of axes.
    Each axis corresponds to a bundle.
    Axis placement should roughly correspond
    to the actual bundle placement in the brain.
    '''

    def __init__(self, size=(5, 5), positions=vut.POSITIONS, fig=None):
        self.size = size
        self.positions = positions
        self.on_grid = np.zeros((5, 5), dtype=bool)

        if fig is None:
            self.fig = plt.figure()
            label = "1"
            self.twinning = False
        else:  # we are twinning with another BrainAxes
            self.fig = fig
            label = "2"
            self.twinning = True
        self.axes = self.fig.subplots(
            self.size[0],
            self.size[1],
            subplot_kw={"label": label, "frame_on": False})
        self.fig.subplots_adjust(
            left=None,
            bottom=None,
            right=None,
            top=None,
            wspace=0.4,
            hspace=0.6)
        self.fig.set_size_inches((18, 18))
        self.temp_fig, self.temp_axis = plt.subplots()
        self.temp_axis_owner = None

    def get_axis(self, bundle, axes_dict={}):
        '''
        Given a bundle, turn on and get an axis.
        If bundle not positioned, will claim the temporary axis.
        If bundle in axes_dict, onyl return relevant axis.
        '''
        if bundle in axes_dict.keys():
            return axes_dict[bundle]
        elif bundle in self.positions.keys():
            self.on_grid[self.positions[bundle]] = True
            return self.axes[self.positions[bundle]]
        else:
            if self.temp_axis_owner != bundle:
                plt.close(self.temp_fig)
                self.temp_fig, self.temp_axis = plt.subplots()
                self.temp_axis_owner = bundle
            return self.temp_axis

    def plot_line(self, bundle, x, y, data, ylabel, ylim, n_boot, alpha,
                  lineplot_kwargs, plot_subject_lines=True, ax=None):
        '''
        Given a dataframe data with at least columns x, y,
        and subjectID, plot the mean of subjects with ci of 95
        in alpha and the individual subjects in alpha-0.2
        using sns.lineplot()
        '''
        if ax is None:
            ax = self.get_axis(bundle)
        lineplot_kwargs_mean = lineplot_kwargs.copy()
        sns.set(style="whitegrid", rc={"lines.linewidth": 1})
        ax.hlines(0, 0, 95, linestyles='dashed', color='red')
        if plot_subject_lines:
            lineplot_kwargs_mean["hue"] = None
            lineplot_kwargs_mean["palette"] = None
            lineplot_kwargs_mean["color"] = "lightgray"

            sns.set(style="whitegrid", rc={"lines.linewidth": 0.5})
            sns.lineplot(
                x=x, y=y,
                data=data,
                ci=None, estimator=None, units='subjectID',
                legend=False, ax=ax, alpha=alpha - 0.2,
                style=[True] * len(data.index), **lineplot_kwargs)

        sns.set(style="whitegrid", rc={"lines.linewidth": 3})
        sns.lineplot(
            x=x, y=y,
            data=data,
            estimator='mean',
            errorbar=('ci', 95),
            n_boot=n_boot,
            legend=False, ax=ax, alpha=alpha,
            style=[True] * len(data.index), **lineplot_kwargs_mean)

        ax.set_title(display_string(bundle), fontsize=vut.large_font)
        ax.set_ylabel(ylabel, fontsize=vut.medium_font)
        ax.set_ylim(ylim)

    def format(self, disable_x=True, disable_y=True):
        '''
        Call this functions once after all axes that you intend to use
        have been plotted on. Automatically formats brain axes.
        '''
        for i in range(self.size[0]):
            for j in range(self.size[1]):
                if self.twinning:
                    self.axes[i, j].xaxis.tick_top()
                    self.axes[i, j].yaxis.tick_right()
                    self.axes[i, j].xaxis.set_label_position('top')
                    self.axes[i, j].yaxis.set_label_position('right')
                self.axes[i, j].tick_params(
                    axis='y', which='major', labelsize=vut.small_font)
                self.axes[i, j].tick_params(
                    axis='x', which='major', labelsize=vut.small_font)
                if not self.on_grid[i, j]:
                    self.axes[i, j].axis("off")
                if self.twinning:
                    if j != self.size[1] - 1 and self.on_grid[i][j + 1]:
                        self.axes[i, j].set_yticklabels([])
                        self.axes[i, j].set_ylabel("")
                    self.axes[i, j].set_xticklabels([])
                    self.axes[i, j].set_xlabel("")
                else:
                    if disable_y and (j != 0 and self.on_grid[i][j - 1]):
                        self.axes[i, j].set_yticklabels([])
                        self.axes[i, j].set_ylabel("")
                    if disable_x or (i != self.size[0] - 1
                                     and self.on_grid[i + 1][j]):
                        self.axes[i, j].set_xticklabels([])
                        self.axes[i, j].set_xlabel("")
        self.fig.tight_layout()

    def save_temp_fig(self, o_folder, o_file, save_func):
        '''
        If using a temporary axis, save it out and clear it.
        Returns True if temporary axis was saved, false if no
        temporary axis was in use
        '''
        if self.temp_axis_owner is None:
            return False
        self.temp_fig.tight_layout()
        save_func(self.temp_fig, o_folder, o_file)
        plt.close(self.temp_fig)
        self.temp_axis_owner = None
        return True

    def is_using_temp_axis(self):
        return (self.temp_axis_owner is not None)

    def close_all(self):
        '''
        Close all associated figures.
        '''
        plt.close(self.temp_fig)
        plt.close(self.fig)


class GroupCSVComparison():
    """
    Compare different CSVs, using:
    tract profiles, contrast indices,
    scan-rescan reliability using ICC.
    """

    def __init__(self, out_folder, csv_fnames, names, is_special="",
                 subjects=None,
                 scalar_bounds={'lb': {'FA': 0.2},
                                'ub': {'MD': 0.002}},
                 bundles=None,
                 percent_nan_tol=10,
                 percent_edges_removed=10,
                 remove_model=True,
                 mat_bundle_converter=BUNDLE_MAT_2_PYTHON,
                 mat_column_converter=vut.CSV_MAT_2_PYTHON,
                 mat_scale_converter=vut.SCALE_MAT_2_PYTHON,
                 bundle_converter=BUNDLE_RECO_2_AFQ,
                 ICC_func="ICC2"):
        """
        Load in csv files, converting from matlab if necessary.

        Parameters
        ----------
        out_folder : path
            Folder where outputs of this class's methods will be saved.

        csv_fnames : list of filenames
            Filenames for the two CSVs containing tract profiles to compare.
            Will obtain subject list from the first file.

        names : list of strings
            Name to use to identify each CSV dataset.

        is_special : str or list of strs, optional
            Whether or not the csv needs special attention.
            Can be "", "mat" if the csv was generated using mAFQ,
            or "reco" if the csv was generated using Recobundles.
            Default: ""

        subjects : list of num, optional
            List of subjects to consider.
            If None, will use all subjects in first dataset.
            Default: None

        scalar_bounds : dictionary, optional
            A dictionary with a lower bound and upper bound containting a
            series of scalar / threshold pairs used as a white-matter mask
            on the profiles (any values outside of the threshold will be
            marked NaN and not used or set to 0, depending on the case).
            Default: {'lb': {'FA': 0.2}, 'ub': {'MD': 0.002}}

        bundles : list of strings, optional
            Bundles to compare.
            If None, use all bundles in the first profile group.
            Default: None

        percent_nan_tol : int, optional
            Percentage of NaNs tolerable. If a profile has less than this
            percentage of NaNs, NaNs are interpolated. If it has more,
            the profile is thrown out.
            Default: 10

        percent_edges_removed : int, optional
            Percentage of nodes to remove from the edges of the profile.
            Scalar values often change dramatically at the boundary between
            the grey matter and the white matter, and these effects can
            dominate plots. However, they are generally not interesting to us,
            and have low intersubject reliability.
            In a profile of 100 nodes, percent_edges_removed=10 would remove
            5 nodes from each edge.
            Default: 10

        remove_model : bool, optional
            Whether to remove prefix of scalars which specify model
            i.e., dti_fa => FA.
            Default: True

        mat_bundle_converter : dictionary, optional
            Dictionary that maps matlab bundle names to python bundle names.
            Default: BUNDLE_MAT_2_PYTHON

        mat_column_converter : dictionary, optional
            Dictionary that maps matlab column names to python column names.
            Default: CSV_MAT_2_PYTHON

        mat_scale_converter : dictionary, optional
            Dictionary that maps scalar names to how they should be scaled
            to match pyAFQ's scale for that scalar.
            Default: SCALE_MAT_2_PYTHON

        bundle_converter : dictionary, optional
            Dictionary that maps bundle names to more standard bundle names.
            Unlike mat_bundle_converter, this converter is applied to all CSVs
            Default: BUNDLE_RECO_2_AFQ

        ICC_func : string, optional
            ICC function to use to calculate correlations.
            Can be 'ICC1, 'ICC2', 'ICC3', 'ICC1k', 'ICC2k', 'ICC3k'.
            Default: "ICC2"
        """
        self.logger = logging.getLogger('AFQ')
        self.ICC_func = ICC_func
        if "k" in self.ICC_func:
            self.ICC_func_name = f"ICC({self.ICC_func[3]},k)"
        else:
            self.ICC_func_name = f"ICC({self.ICC_func[3]},1)"
        self.out_folder = out_folder
        self.percent_nan_tol = percent_nan_tol

        if not isinstance(is_special, list):
            is_special = [is_special] * len(csv_fnames)

        self.profile_dict = {}
        for i, fname in enumerate(csv_fnames):
            profile = pd.read_csv(fname)
            if 'subjectID' in profile.columns:
                profile['subjectID'] = \
                    profile['subjectID'].apply(
                        lambda x: int(
                            ''.join(c for c in x if c.isdigit())
                        ) if isinstance(x, str) else x)
            else:
                profile['subjectID'] = 0

            if is_special[i] == "mat":
                profile.rename(
                    columns=mat_column_converter, inplace=True)
                profile['tractID'] = \
                    profile['tractID'].apply(
                        lambda x: mat_bundle_converter[x])
                for scalar, scale in mat_scale_converter.items():
                    profile[scalar] = \
                        profile[scalar].apply(lambda x: x * scale)
            profile.replace({"tractID": bundle_converter}, inplace=True)
            if is_special[i] == "reco":
                def reco_flip(df):
                    if df.tractID in vut.RECO_FLIP:
                        return 99 - df.nodeID
                    else:
                        return df.nodeID
                profile["nodeID"] = profile.apply(reco_flip, axis=1)
            if remove_model:
                profile.rename(
                    columns=vut.SCALAR_REMOVE_MODEL, inplace=True)

            for bound, constraint in scalar_bounds.items():
                for scalar, threshold in constraint.items():
                    profile[scalar] = \
                        profile[scalar].apply(
                            lambda x: self._threshold_scalar(
                                bound,
                                threshold,
                                x))

            if percent_edges_removed > 0:
                profile = profile.drop(profile[np.logical_or(
                    (profile["nodeID"] < percent_nan_tol // 2),
                    (profile["nodeID"] >= 100 - (percent_nan_tol // 2))
                )].index)

            self.profile_dict[names[i]] = profile
        if subjects is None:
            self.subjects = self.profile_dict[names[0]]['subjectID'].unique()
        else:
            self.subjects = subjects
        self.prof_len = 100 - (percent_nan_tol // 2) * 2
        if bundles is None:
            self.bundles = self.profile_dict[names[0]]['tractID'].unique()
            self.bundles.sort()
        else:
            self.bundles = bundles
        self.color_dict = vut.gen_color_dict([*self.bundles, "median"])

        # TODO: make these parameters
        self.scalar_markers = ["o", "x"]
        self.patterns = (
            None, '/', 'o', 'x', '-', '.',
            '+', '//', '\\', '*', 'O', '|')

    def _threshold_scalar(self, bound, threshold, val):
        """
        Threshold scalars by a lower and upper bound.
        """
        if bound == "lb":
            if val > threshold:
                return val
            else:
                return np.nan
        elif bound == "ub":
            if val < threshold:
                return val
            else:
                return np.nan
        else:
            raise RuntimeError("scalar_bounds dictionary "
                               + " formatted incorrectly. See"
                               + " the default for reference")

    def _save_fig(self, fig, folder, f_name):
        """
        Get file to save to, and generate the folder if it does not exist.
        """
        if self.out_folder is None and folder is None:
            f_folder = None
        elif self.out_folder is None:
            f_folder = folder
        elif folder is None:
            f_folder = self.out_folder
        else:
            f_folder = op.join(
                self.out_folder,
                folder)

        if f_folder is None:
            fig.savefig(f_name + ".png")
            fig.savefig(f_name + ".svg",
                        format='svg',
                        dpi=300)
        else:
            os.makedirs(f_folder, exist_ok=True)
            fig.savefig(op.join(f_folder, f_name) + ".png")
            fig.savefig(op.join(f_folder, f_name) + ".svg",
                        format='svg',
                        dpi=300)

    def _get_profile(self, name, bundle, subject, scalar):
        """
        Get a single profile, then handle not found / NaNs
        """
        profile = self.profile_dict[name]
        single_profile = profile[
            (profile['subjectID'] == subject)
            & (profile['tractID'] == bundle)
        ].sort_values("nodeID")[scalar].to_numpy()
        nans = np.isnan(single_profile)
        percent_nan = (np.sum(nans) * 100) // self.prof_len
        if len(single_profile) < 1:
            self.logger.warning(
                'No scalars found for scalar ' + scalar
                + ' for subject ' + str(subject)
                + ' for bundle ' + bundle
                + ' for CSV ' + name)
            return None

        if percent_nan > 0:
            message = (
                f'{percent_nan}% NaNs found in scalar ' + scalar
                + ' for subject ' + str(subject)
                + ' for bundle ' + bundle
                + ' for CSV ' + name)
            if np.sum(nans) > self.percent_nan_tol:
                self.logger.warning(message + '. Profile ignored. ')
                return None
            else:
                self.logger.info(message + '. NaNs interpolated. ')
                non_nan = np.logical_not(nans)
                single_profile[nans] = np.interp(
                    nans.nonzero()[0],
                    non_nan.nonzero()[0],
                    single_profile[non_nan])

        return single_profile

    def _alpha(self, alpha):
        '''
        Keep alpha in a reasonable range
        Useful when calculating alpha automatically
        '''
        if alpha < 0.3:
            return 0.3
        if alpha > 1:
            return 1
        return alpha

    def _array_to_df(self, arr):
        '''
        Converts a 2xn array to a pandas dataframe with columns x, y
        Useful for plotting with seaborn.
        '''
        df = pd.DataFrame()
        df['x'] = arr[0]
        df['y'] = arr[1]
        return df

    def masked_corr(self, arr, corrtype):
        '''
        Mask arr for NaNs before calling np.corrcoef
        '''
        mask = np.logical_not(
            np.logical_or(
                np.isnan(arr[0, ...]),
                np.isnan(arr[1, ...])))
        if np.sum(mask) < 2:
            return np.nan, np.nan, np.nan
        arr = arr[:, mask]
        if corrtype == "ICC":
            data = pd.DataFrame({
                "targets": np.concatenate(
                    (np.arange(arr.shape[1]), np.arange(arr.shape[1]))),
                "raters": np.concatenate(
                    (np.zeros(arr.shape[1]), np.ones(arr.shape[1]))),
                "ratings": np.concatenate(
                    (arr[0], arr[1]))})
            stats = intraclass_corr(
                data=data,
                targets="targets",
                raters="raters",
                ratings="ratings")
            row = stats[stats["Type"] == self.ICC_func].iloc[0]
            return row["ICC"], row["ICC"] - row["CI95%"][0], \
                row["CI95%"][1] - row["ICC"]
        elif corrtype == "Srho":
            stats = corr(
                x=arr[0],
                y=arr[1],
                method="spearman")
            row = stats.iloc[0]
            return row["r"], row["r"] - row["CI95%"][0], \
                row["CI95%"][1] - row["r"]
        else:
            raise ValueError("corrtype not recognized")

    def tract_profiles(self, names=None, scalar="FA",
                       ylim=[0.0, 1.0],
                       show_plots=False,
                       positions=vut.POSITIONS,
                       out_file=None,
                       n_boot=1000,
                       plot_subject_lines=True,
                       axes_dict={}):
        """
        Compare all tract profiles for a scalar from different CSVs.
        Plots tract profiles for all in one plot.
        Bundles taken from positions argument.

        Parameters
        ----------
        names : list of strings, optional
            Names of datasets to plot profiles of.
            If None, all datasets are used.
            Default: None

        scalar : string, optional
            Scalar to use in plots. Default: "FA".

        ylim : list of 2 floats, optional
            Minimum and maximum value used for y-axis bounds.
            If None, ylim is not set.
            Default: [0.0, 1.0]

        out_file : str, optional
            Path to save the figure to.
            If None, use the default naming convention in self.out_folder
            Default: None

        n_boot : int, optional
            Number of bootstrap resamples for seaborn to use
            to estimate the ci.
            Default: 1000

        show_plots : bool, optional
            Whether to show plots if in an interactive environment.
            Default: False

        positions : dictionary, optional
            Dictionary that maps bundle names to position in plot.
            Default: vut.POSITIONS

        plot_subject_lines : bool, optional
            Whether to plot individual subject lines with a smaller width.
            Default: True

        axes_dict : dictionary of axes, optional
            Plot contrast index for bundles that are keys of
            axes_dict on the corresponding axis. Default: {}
        """
        if not show_plots:
            plt.ioff()
        if names is None:
            names = list(self.profile_dict.keys())
        if out_file is None:
            o_folder = f"tract_profiles/{scalar}"
            o_file = f"{'_'.join(names)}"
        else:
            o_folder = None
            o_file = out_file

        ba = BrainAxes(positions=positions)
        labels = []
        self.logger.info("Calculating means and CIs...")
        for j, bundle in enumerate(tqdm(self.bundles)):
            labels_temp = []
            for i, name in enumerate(names):
                if i == 0:
                    plot_kwargs = {
                        "hue": "tractID",
                        "palette": [self.color_dict[bundle]]}
                else:
                    plot_kwargs = {
                        "dashes": [(2**(i - 1), 2**(i - 1))],
                        "hue": "tractID",
                        "palette": [self.color_dict[bundle]]}
                profile = self.profile_dict[name]
                profile = profile[profile['tractID'] == bundle]
                ba.plot_line(
                    bundle, "nodeID", scalar, profile,
                    display_string(scalar), ylim, n_boot, self._alpha(
                        0.6 + 0.2 * i),
                    plot_kwargs,
                    plot_subject_lines=plot_subject_lines,
                    ax=axes_dict.get(bundle))
                if j == 0:
                    line = Line2D(
                        [], [],
                        color=[0, 0, 0])
                    line.set_dashes((2**(i + 1), 2**(i + 1)))
                    if ba.is_using_temp_axis():
                        labels_temp.append(line)
                    else:
                        labels.append(line)
            if ba.is_using_temp_axis():
                ba.temp_fig.legend(
                    labels_temp, names, fontsize=vut.medium_font)
                ba.save_temp_fig(
                    o_folder, f"{o_file}_{bundle}", self._save_fig)
        if len(names) > 1:
            ba.fig.legend(
                labels, names, loc='center',
                fontsize=vut.medium_font)
        ba.format()

        self._save_fig(ba.fig, o_folder, o_file)

        if not show_plots:
            ba.close_all()
            plt.ion()

    def _contrast_index_df_maker(self, bundles, names, scalar):
        ci_df = pd.DataFrame(columns=["subjectID", "nodeID", "diff"])
        for subject in self.subjects:
            profiles = [None] * 2
            both_found = True
            for i, name in enumerate(names):
                for j, bundle in enumerate(bundles):
                    profiles[i + j] = self._get_profile(
                        name, bundle, subject, scalar)
                    if profiles[i + j] is None:
                        both_found = False
            if both_found:
                this_contrast_index = \
                    calc_contrast_index(profiles[0], profiles[1])
                for i, diff in enumerate(this_contrast_index):
                    ci_df = ci_df.append({
                        "subjectID": subject,
                        "nodeID": i,
                        "diff": diff},
                        ignore_index=True)
        return ci_df

    def contrast_index(self, names=None, scalar="FA",
                       show_plots=False, n_boot=1000,
                       ylim=(-0.5, 0.5), show_legend=False,
                       positions=vut.POSITIONS, plot_subject_lines=True,
                       axes_dict={}):
        """
        Calculate the contrast index for each bundle in two datasets.

        Parameters
        ----------
        names : list of strings, optional
            Names of datasets to plot profiles of.
            If None, all datasets are used.
            Should be a total of only two datasets.
            Default: None

        scalar : string, optional
            Scalar to use for the contrast index. Default: "FA".

        show_plots : bool, optional
            Whether to show plots if in an interactive environment.
            Default: False

        n_boot : int, optional
            Number of bootstrap resamples for seaborn to use
            to estimate the ci.
            Default: 1000

        ylim : list of 2 floats, optional
            Minimum and maximum value used for y-axis bounds.
            If None, ylim is not set.
            Default: None

        show_legend : bool, optional
            Show legend in center with single entry denoting the scalar used.
            Default: False

        positions : dictionary, optional
            Dictionary that maps bundle names to position in plot.
            Default: vut.POSITIONS

        plot_subject_lines : bool, optional
            Whether to plot individual subject lines with a smaller width.
            Default: True

        axes_dict : dictionary of axes, optional
            Plot contrast index for bundles that are keys of
            axes_dict on the corresponding axis. Default: {}
        """
        if not show_plots:
            plt.ioff()

        if names is None:
            names = list(self.profile_dict.keys())
        if len(names) != 2:
            self.logger.error("To calculate the contrast index, "
                              + "only two dataset names should be given")
            return None

        ba = BrainAxes(positions=positions)
        ci_all_df = {}
        for j, bundle in enumerate(tqdm(self.bundles)):
            ci_df = self._contrast_index_df_maker(
                [bundle], names, scalar)
            ba.plot_line(
                bundle, "nodeID", "diff", ci_df, "ACI", ylim,
                n_boot, 1.0, {"color": self.color_dict[bundle]},
                plot_subject_lines=plot_subject_lines,
                ax=axes_dict.get(bundle))
            ci_all_df[bundle] = ci_df
            ba.save_temp_fig(
                f"contrast_plots/{scalar}/",
                f"{names[0]}_vs_{names[1]}_contrast_index_{bundle}",
                self._save_fig)
        if show_legend:
            ba.fig.legend([scalar], loc='center', fontsize=vut.medium_font)
        ba.format()
        self._save_fig(
            ba.fig,
            f"contrast_plots/{scalar}/",
            f"{names[0]}_vs_{names[1]}_contrast_index")
        if not show_plots:
            ba.close_all()
            plt.ion()
        return ba.fig, ba.axes, ci_all_df

    def lateral_contrast_index(self, name, scalar="FA",
                               show_plots=False, n_boot=1000,
                               ylim=(-1, 1),
                               positions=vut.POSITIONS,
                               plot_subject_lines=True):
        """
        Calculate the lateral contrast index for each bundle in a given
        dataset, for each dataset in names.

        Parameters
        ----------
        name : string
            Names of dataset to plot profiles of.

        scalar : string, optional
            Scalar to use for the contrast index. Default: "FA".

        show_plots : bool, optional
            Whether to show plots if in an interactive environment.
            Default: False

        n_boot : int, optional
            Number of bootstrap resamples for seaborn to use
            to estimate the ci.
            Default: 1000

        ylim : list of 2 floats, optional
            Minimum and maximum value used for y-axis bounds.
            If None, ylim is not set.
            Default: None

        positions : dictionary, optional
            Dictionary that maps bundle names to position in plot.
            Default: vut.POSITIONS

        plot_subject_lines : bool, optional
            Whether to plot individual subject lines with a smaller width.
            Default: True
        """
        if not show_plots:
            plt.ioff()

        ba = BrainAxes(positions=positions)
        for j, bundle in enumerate(tqdm(self.bundles)):
            other_bundle = list(bundle)
            if other_bundle[-1] == 'L':
                other_bundle[-1] = 'R'
            elif other_bundle[-1] == 'R':
                other_bundle[-1] = 'L'
            else:
                continue
            other_bundle = "".join(other_bundle)
            if other_bundle not in self.bundles:
                continue

            ci_df = self._contrast_index_df_maker(
                [bundle, other_bundle], [name], scalar)
            ba.plot_line(
                bundle, "nodeID", "diff", ci_df, "ACI", ylim,
                n_boot, 1.0, {"color": self.color_dict[bundle]},
                plot_subject_lines=plot_subject_lines)
            ba.save_temp_fig(
                f"contrast_plots/{scalar}/",
                f"{name}_lateral_contrast_index_{bundle}",
                self._save_fig)

        ba.fig.legend([scalar], loc='center', fontsize=vut.medium_font)
        ba.format()
        self._save_fig(
            ba.fig,
            f"contrast_plots/{scalar}/",
            f"{name}_lateral_contrast_index")
        if not show_plots:
            ba.close_all()

        if not show_plots:
            plt.ion()

    def reliability_plots(self, names=None,
                          scalars=["FA", "MD"],
                          ylims=[0.0, 1.0],
                          show_plots=False,
                          only_plot_above_thr=None,
                          rotate_y_labels=False,
                          rtype="Reliability",
                          positions=vut.POSITIONS,
                          fig_axes=None,
                          prof_axes_dict={},
                          sub_axes_dict={}):
        """
        Plot the scan-rescan reliability using ICC for 2 scalars.

        Parameters
        ----------
        names : list of strings, optional
            Names of datasets to plot profiles of.
            If None, all datasets are used.
            Should be a total of only two datasets.
            Default: None

        scalars : list of strings, optional
            Scalars to correlate. Default: ["FA", "MD"].

        ylims : 2-tuple of floats, optional
            Limits of the y-axis. Useful to synchronize axes across graphs.
            Default: [0.0, 1.0].

        show_plots : bool, optional
            Whether to show plots if in an interactive environment.
            Default: False

        only_plot_above_thr : int or None, optional
            Only plot bundles with intrersubject reliability above this
            threshold on the final reliability bar plots. If None, plot all.
            Default: None

        rotate_y_labels : bool, optional
            Rotate y labels on final reliability plots.
            Default: False

        rtype : str, optional
            Type of reliability to name the y axis of the reliability bar
            charts. Default: "Reliability"

        positions : dictionary, optional
            Dictionary that maps bundle names to position in plot.
            Default: vut.POSITIONS

        fig_axes : tuple of matplotlib figure and axes, optional
            If not None, the resulting reliability plots will use this
            figure and axes. Default: None

        prof_axes_dict : dictionary of axes, optional
            Plot profile reliability histograms for bundles that
            are keys of prof_axes_dict on the corresponding axis.
            Default: {}

        sub_axes_dict : dictionary of axes, optional
            Plot subject reliability scatter plots for bundles that
            are keys of sub_axes_dict on the corresponding axis.
            Default: {}

        Returns
        -------
        Returns 8 objects:
        1. Matplotlib figure
        2. Matplotlib axes
        3. A dictionary containing the number of missing bundles
           for each dataset.
        4. A list of bundles with sufficient correlation
        5. A pandas dataframe describing the intersubject reliabilities,
           per bundle
        6. A numpy array describing the intersubject reliability errors,
           per bundle
        7. A pandas dataframe desribing the profile reliabilities,
           per bundle
        8. A numpy array desribing the profile reliability errors,
           per bundle
        """
        if not show_plots:
            plt.ioff()
        if names is None:
            names = list(self.profile_dict.keys())
        if len(names) != 2:
            self.logger.error("To plot correlations, "
                              + "only two dataset names should be given")
            return None

        # extract relevant statistics / data from profiles
        N = len(self.subjects)
        all_sub_coef = np.zeros((len(scalars), len(self.bundles)))
        all_sub_coef_err = np.zeros((len(scalars), len(self.bundles), 2))
        all_sub_means = np.zeros(
            (len(scalars), len(self.bundles), 2, N))
        all_profile_coef = \
            np.zeros((len(scalars), len(self.bundles), N))
        all_node_coef = np.zeros(
            (len(scalars), len(self.bundles), self.prof_len))
        miss_counts = pd.DataFrame(0, index=self.bundles, columns=[
            f"miss_count{names[0]}", f"miss_count{names[1]}"])
        for m, scalar in enumerate(scalars):
            for k, bundle in enumerate(tqdm(self.bundles)):
                bundle_profiles =\
                    np.zeros((2, N, self.prof_len))
                for j, name in enumerate(names):
                    for i, subject in enumerate(self.subjects):
                        single_profile = self._get_profile(
                            name, bundle, subject, scalar)
                        if single_profile is None:
                            bundle_profiles[j, i] = np.nan
                            miss_counts.at[bundle, f"miss_count{name}"] =\
                                miss_counts.at[
                                    bundle, f"miss_count{name}"] + 1
                        else:
                            bundle_profiles[j, i] = single_profile

                all_sub_means[m, k] = np.nanmean(bundle_profiles, axis=2)
                all_sub_coef[m, k], all_sub_coef_err[m, k, 0], \
                    all_sub_coef_err[m, k, 1] =\
                    self.masked_corr(all_sub_means[m, k], "Srho")
                if np.isnan(all_sub_coef[m, k]).all():
                    self.logger.error((
                        f"Not enough non-nan profiles"
                        f"for scalar {scalar} for bundle {bundle}"))
                    all_sub_coef[m, k] = 0

                bundle_coefs = np.zeros(N)
                for i in range(N):
                    bundle_coefs[i], _, _ = \
                        self.masked_corr(bundle_profiles[:, i, :], "ICC")
                all_profile_coef[m, k] = bundle_coefs

                node_coefs = np.zeros(self.prof_len)
                for i in range(self.prof_len):
                    node_coefs[i], _, _ =\
                        self.masked_corr(bundle_profiles[:, :, i], "ICC")
                all_node_coef[m, k] = node_coefs

        # plot histograms of subject ICC
        maxi = np.nanmax(all_profile_coef)
        mini = np.nanmin(all_profile_coef)
        bins = np.linspace(mini, maxi, 10)
        ba = BrainAxes(positions=positions)
        for k, bundle in enumerate(self.bundles):
            ax = ba.get_axis(bundle, axes_dict=prof_axes_dict)
            for m, scalar in enumerate(scalars):
                bundle_coefs = all_profile_coef[m, k]
                bundle_coefs = bundle_coefs[~np.isnan(bundle_coefs)]
                sns.set(style="whitegrid")
                sns.histplot(
                    data=bundle_coefs,
                    bins=bins,
                    alpha=0.5,
                    color=self.color_dict[bundle],
                    hatch=self.patterns[m],
                    label=scalar,
                    ax=ax)
            ax.set_title(display_string(bundle), fontsize=vut.large_font)
            ax.set_xlabel(self.ICC_func_name, fontsize=vut.medium_font)
            ax.set_ylabel("Subject count", fontsize=vut.medium_font)
            ba.temp_fig.legend(
                display_string(scalars),
                fontsize=vut.medium_font)
            ba.save_temp_fig(
                f"rel_plots/{'_'.join(scalars)}/verbose",
                (f"{names[0]}_vs_{names[1]}_profile_r_distributions"
                 f"_{bundle}"),
                self._save_fig)

        legend_labels = []
        for m, _ in enumerate(scalars):
            legend_labels.append(Patch(
                facecolor='k',
                hatch=self.patterns[m]))
        ba.fig.legend(
            legend_labels, display_string(scalars),
            loc='center', fontsize=vut.medium_font)
        ba.format(disable_x=False)
        self._save_fig(
            ba.fig,
            f"rel_plots/{'_'.join(scalars)}/verbose",
            f"{names[0]}_vs_{names[1]}_profile_r_distributions")

        if not show_plots:
            ba.close_all()

        # plot node reliability profile
        all_node_coef[np.isnan(all_node_coef)] = 0
        if ylims is None:
            maxi = all_node_coef.max()
            mini = all_node_coef.min()
        else:
            maxi = ylims[1]
            mini = ylims[0]
        ba = BrainAxes(positions=positions)
        for k, bundle in enumerate(self.bundles):
            ax = ba.get_axis(bundle)
            for m, scalar in enumerate(scalars):
                sns.set(style="whitegrid")
                sns.lineplot(
                    data=all_node_coef[m, k],
                    label=scalar,
                    color=vut.tableau_20[m * 2],
                    ax=ax,
                    legend=False,
                    ci=None, estimator=None)
            ax.set_ylim([mini, maxi])
            ax.set_title(display_string(bundle), fontsize=vut.large_font)
            ax.set_ylabel(self.ICC_func_name, fontsize=vut.medium_font)
            ba.temp_fig.legend(
                display_string(scalars),
                fontsize=vut.medium_font)
            ba.save_temp_fig(
                f"rel_plots/{'_'.join(scalars)}/verbose",
                (f"{names[0]}_vs_{names[1]}_node_profiles"
                    f"_{bundle}"),
                self._save_fig)

        ba.fig.legend(display_string(scalars),
                      loc='center', fontsize=vut.medium_font)
        ba.format()
        self._save_fig(
            ba.fig,
            f"rel_plots/{'_'.join(scalars)}/verbose",
            f"{names[0]}_vs_{names[1]}_node_profiles")

        if not show_plots:
            ba.close_all()

        # plot mean profile scatter plots
        for m, scalar in enumerate(scalars):
            maxi = np.nanmax(all_sub_means[m])
            mini = np.nanmin(all_sub_means[m])
            if len(scalars) == 2:
                twinning_next = (m == 0)
                twinning = (m == 1)
            else:
                twinning = False
                twinning_next = False
            if twinning:
                ba = BrainAxes(positions=positions, fig=ba.fig)
            else:
                ba = BrainAxes(positions=positions)
            for k, bundle in enumerate(self.bundles):
                if twinning:
                    fc = 'w'
                    ec = self.color_dict[bundle]
                else:
                    fc = self.color_dict[bundle]
                    ec = 'w'
                ax = ba.get_axis(bundle, axes_dict=sub_axes_dict)
                sns.set(style="whitegrid")
                if not twinning:
                    ax.plot(
                        [[0, 0], [1, 1]], [[0, 0], [1, 1]],
                        '--', color='red')
                ax.scatter(
                    all_sub_means[m, k, 0],
                    all_sub_means[m, k, 1],
                    label=scalar,
                    marker=self.scalar_markers[m - twinning],
                    facecolors=fc,
                    edgecolors=ec,
                    s=vut.marker_size,
                    linewidth=1)
                if twinning or twinning_next:
                    twinning_color = 'k'
                    if twinning_next:
                        ax.spines['bottom'].set_color(twinning_color)
                        ax.spines['left'].set_color(twinning_color)
                    else:
                        ax.spines['top'].set_color(twinning_color)
                        ax.spines['right'].set_color(twinning_color)
                    ax.xaxis.label.set_color(twinning_color)
                    ax.tick_params(axis='x', colors=twinning_color)
                    ax.xaxis.label.set_color(twinning_color)
                    ax.yaxis.label.set_color(twinning_color)
                    ax.tick_params(axis='y', colors=twinning_color)
                    ax.yaxis.label.set_color(twinning_color)
                if not twinning:
                    ax.set_title(
                        display_string(bundle),
                        fontsize=vut.large_font)
                ax.set_xlabel(names[0], fontsize=vut.medium_font)
                ax.set_ylabel(names[1], fontsize=vut.medium_font)
                ax.set_ylim([mini, maxi])
                ax.set_xlim([mini, maxi])
                ba.temp_fig.legend(
                    [scalar], fontsize=vut.medium_font)
                ba.save_temp_fig(
                    f"rel_plots/{'_'.join(scalars)}/verbose",
                    (f"{names[0]}_vs_{names[1]}_{scalar}_mean_profiles"
                        f"_{bundle}"),
                    self._save_fig)
            if twinning:
                legend_labels = [
                    Line2D(
                        [], [],
                        markerfacecolor='k',
                        markeredgecolor='w',
                        marker=self.scalar_markers[0],
                        linewidth=0,
                        markersize=15),
                    Line2D(
                        [], [],
                        markeredgecolor='k',
                        markerfacecolor='w',
                        marker=self.scalar_markers[0],
                        linewidth=0,
                        markersize=15)]
                ba.fig.legend(
                    legend_labels,
                    display_string(scalars),
                    loc='center',
                    fontsize=vut.medium_font)
            elif not twinning_next:
                ba.fig.legend([scalar], loc='center', fontsize=vut.medium_font)
                self._save_fig(
                    ba.fig,
                    f"rel_plots/{'_'.join(scalars)}/verbose",
                    f"{names[0]}_vs_{names[1]}_{scalar}_mean_profiles")
            ba.format(disable_x=False)
            if not (show_plots or twinning_next):
                ba.close_all()

        # plot bar plots of ICC
        if fig_axes is None:
            fig, axes = plt.subplots(2, 1)
            fig.set_size_inches((8, 8))
        else:
            fig = fig_axes[0]
            axes = fig_axes[1]
        bundle_prof_means = np.nanmean(all_profile_coef, axis=2)
        bundle_prof_stds = 1.95 * \
            sem(all_profile_coef, axis=2, nan_policy='omit')
        if ylims is None:
            maxi = np.maximum(bundle_prof_means.max(), all_sub_coef.max())
            mini = np.minimum(bundle_prof_means.min(), all_sub_coef.min())
        else:
            maxi = ylims[1]
            mini = ylims[0]

        if only_plot_above_thr is not None:
            is_removed_bundle =\
                np.logical_not(
                    np.logical_and(
                        np.all(all_sub_coef > only_plot_above_thr, axis=0),
                        np.all(bundle_prof_means > only_plot_above_thr,
                               axis=0)))
            removal_idx = np.where(is_removed_bundle)[0]
            bundle_prof_means_removed = np.delete(
                bundle_prof_means,
                removal_idx,
                axis=1)
            bundle_prof_stds_removed = np.delete(
                bundle_prof_stds,
                removal_idx,
                axis=1)
            all_sub_coef_removed = np.delete(
                all_sub_coef,
                removal_idx,
                axis=1)
            all_sub_coef_err_removed = np.delete(
                all_sub_coef_err,
                removal_idx,
                axis=1)
        else:
            is_removed_bundle = [False] * len(self.bundles)
            bundle_prof_means_removed = bundle_prof_means
            bundle_prof_stds_removed = bundle_prof_stds
            all_sub_coef_err_removed = all_sub_coef_err
            all_sub_coef_removed = all_sub_coef

        updated_bundles = []
        for k, bundle in enumerate(self.bundles):
            if not is_removed_bundle[k]:
                if bundle == "CC_ForcepsMinor":
                    updated_bundles.append("CC_FMi")
                else:
                    updated_bundles.append(bundle)
        updated_bundles.append("median")

        sns.set(style="whitegrid")
        width = 0.6
        spacing = 1.5
        x = np.arange(len(updated_bundles)) * spacing
        x_shift = np.linspace(-0.5 * width, 0.5 * width, num=len(scalars))

        bundle_prof_means_removed = np.pad(
            bundle_prof_means_removed, [(0, 0), (0, 1)])
        bundle_prof_stds_removed = np.pad(
            bundle_prof_stds_removed, [(0, 0), (0, 1)])
        all_sub_coef_removed = np.pad(
            all_sub_coef_removed, [(0, 0), (0, 1)])
        all_sub_coef_err_removed = np.transpose(np.pad(
            all_sub_coef_err_removed, [(0, 0), (0, 1), (0, 0)]))

        for m, scalar in enumerate(scalars):
            bundle_prof_means_removed[m, -1] = np.median(
                bundle_prof_means_removed[m, :-1])
            all_sub_coef_removed[m, -1] = np.median(
                all_sub_coef_removed[m, :-1])

            # This code can be used as a baseline to make violin plots
            #
            # mask = ~np.isnan(all_profile_coef[m].T)
            # all_profile_coef_m_removed =\
            #     [d[k] for d, k in zip(all_profile_coef[m], mask.T)]
            # vl_parts = axes[0].violinplot(
            #     all_profile_coef_m_removed,
            #     positions=x[:-1] + x_shift[m],
            #     showmedians=True
            #     )
            # color_list = list(self.color_dict.values())
            # for c, pc in enumerate(vl_parts['bodies']):
            #     pc.set_facecolor(color_list[c])
            #     pc.set_edgecolor(color_list[c])
            #     pc.set_alpha(1)
            #     pc.set_hatch(self.patterns[m])
            # vl_parts['cbars'].set_color(color_list)

            axes[0].bar(
                x + x_shift[m],
                bundle_prof_means_removed[m],
                width,
                label=scalar,
                yerr=bundle_prof_stds_removed[m],
                hatch=self.patterns[m],
                color=self.color_dict.values())
            axes[1].bar(
                x + x_shift[m],
                all_sub_coef_removed[m],
                width,
                label=scalar,
                yerr=all_sub_coef_err_removed[:, :, m],
                hatch=self.patterns[m],
                color=self.color_dict.values())

        if len(updated_bundles) > 20:
            xaxis_font_size = vut.small_font - 6
        else:
            xaxis_font_size = vut.small_font

        axes[0].set_title("A", fontsize=vut.large_font)
        axes[0].set_ylabel(f'Profile {rtype}',
                           fontsize=vut.medium_font)
        axes[0].set_ylim([mini, maxi])
        axes[0].set_xlabel("")
        axes[0].set_yticklabels(
            [0, 0.2, 0.4, 0.6, 0.8, 1.0],
            fontsize=vut.small_font - 8)
        axes[0].set_xticks(x + 0.5)
        axes[0].set_xticklabels(
            display_string(updated_bundles), fontsize=xaxis_font_size)
        axes[1].set_title("B", fontsize=vut.large_font)
        axes[1].set_ylabel(f'Subject {rtype}',
                           fontsize=vut.medium_font)
        axes[1].set_ylim([mini, maxi])
        axes[1].set_xlabel("")
        axes[1].set_yticklabels(
            [0, 0.2, 0.4, 0.6, 0.8, 1.0],
            fontsize=vut.small_font - 8)
        axes[1].set_xticks(x + 0.5)
        axes[1].set_xticklabels(
            display_string(updated_bundles), fontsize=xaxis_font_size)

        plt.setp(axes[0].get_xticklabels(),
                 rotation=65,
                 horizontalalignment='right')
        plt.setp(axes[1].get_xticklabels(),
                 rotation=65,
                 horizontalalignment='right')

        if rotate_y_labels:
            plt.setp(axes[0].get_yticklabels(),
                     rotation=90)
            plt.setp(axes[1].get_yticklabels(),
                     rotation=90)

        fig.tight_layout()
        legend_labels = []
        for m, _ in enumerate(scalars):
            legend_labels.append(Patch(
                facecolor='k',
                hatch=self.patterns[m]))
        fig.legend(
            legend_labels,
            display_string(scalars),
            fontsize=vut.small_font,
            bbox_to_anchor=(1.25, 0.5))
        self._save_fig(
            fig,
            f"rel_plots/{'_'.join(scalars)}",
            f"{names[0]}_vs_{names[1]}")

        if not show_plots:
            plt.close(fig)
            plt.ion()
        return fig, axes, miss_counts, updated_bundles, \
            all_sub_coef_removed, all_sub_coef_err_removed, \
            bundle_prof_means_removed, bundle_prof_stds_removed

    def compare_reliability(self, reliability1, reliability2,
                            analysis_label1, analysis_label2,
                            bundles,
                            errors1=None, errors2=None,
                            scalars=["FA", "MD"],
                            rtype="Subject Reliability",
                            show_plots=False,
                            show_legend=True,
                            fig_ax=None):
        """
        Plot a comparison of scan-rescan reliability between two analyses.

        Parameters
        ----------
        reliability1, reliability2 : numpy arrays
            numpy arrays of reliabilities.
            Typically, each of this will be outputs of separate calls
            to reliability_plots.

        analysis_label1, analysis_label2 : Strings
            Names of the analyses used to obtain each dataset.
            Used to label the x and y axes.

        bundles : list of str
            List of bundles that correspond to the second dimension of the
            reliability arrays.

        errors1, errors2 : numpy arrays or None
            Numpy arrays describing the errors.
            Typically, each of this will be outputs of separate calls
            to reliability_plots.
            If None, errors are not shown.
            Default is None.

        scalars : list of str, optional
            Lsit of scalars that correspond to the first dimension of the
            reliability arrays. Default: ["FA", "MD"]

        rtype : str
            type of reliability. Can be any string; used in x axis lavel.
            Default: Subject Reliability

        show_plots : bool, optional
            Whether to show plots if in an interactive environment.
            Default: False

        show_legend : bool, optional
            Show legend for the plot, off to the right hand side.
            Default: True

        fig_ax : tuple of matplotlib figure and axis, optional
            If not None, the resulting reliability plots will use this
            figure and axis. Default: None

        Returns
        -------
        Returns a Matplotlib figure and axes.
        """
        show_error = ((errors1 is not None) and (errors2 is not None))
        if fig_ax is None:
            fig, ax = plt.subplots()
        else:
            fig = fig_ax[0]
            ax = fig_ax[1]
        legend_labels = []
        for i, scalar in enumerate(scalars):
            marker = self.scalar_markers[i]
            if marker == "x":
                marker = marker.upper()
            for j, bundle in enumerate(bundles):
                ax.scatter(
                    reliability1[i, j],
                    reliability2[i, j],
                    s=vut.marker_size,
                    c=[self.color_dict[bundle]],
                    marker=marker
                )
                if show_error:
                    if len(errors1.shape) > 2:
                        xerr = errors1[:, j, i].reshape((2, 1))
                    else:
                        xerr = errors1[i, j]
                    if len(errors2.shape) > 2:
                        yerr = errors2[:, j, i].reshape((2, 1))
                    else:
                        yerr = errors2[i, j]

                    ax.errorbar(
                        reliability1[i, j],
                        reliability2[i, j],
                        xerr=xerr,
                        yerr=yerr,
                        c=[self.color_dict[bundle]],
                        alpha=0.5,
                        fmt="none"
                    )
                if i == 0:
                    legend_labels.append(Patch(
                        facecolor=self.color_dict[bundle],
                        label=bundle))
            legend_labels.append(Line2D(
                [0], [0],
                marker=marker,
                color='k',
                lw=0,
                markersize=10,
                label=scalar))

        ax.set_xlabel(f"{analysis_label1} {rtype}",
                      fontsize=vut.medium_font)
        ax.set_ylabel(f"{analysis_label2} {rtype}",
                      fontsize=vut.medium_font)
        ax.tick_params(
            axis='x', which='major', labelsize=vut.medium_font)
        ax.tick_params(
            axis='y', which='major', labelsize=vut.medium_font)
        ax.set_ylim(0.2, 1)
        ax.set_xlim(0.2, 1)
        ax.plot([[0, 0], [1, 1]], [[0, 0], [1, 1]], '--', color='red')
        legend_labels.append(Line2D(
            [0], [0], linewidth=3, linestyle='--', color='red', label='X=Y'))
        if show_legend:
            fig.legend(
                handles=legend_labels,
                fontsize=vut.small_font - 6,
                bbox_to_anchor=(1.5, 2.0))
            fig.tight_layout()
        return fig, ax


def visualize_gif_inline(fname, use_s3fs=False):
    """Display a gif inline, possible from s3fs """
    if use_s3fs:
        import s3fs
        fs = s3fs.S3FileSystem()
        tdir = tempfile.gettempdir()
        fname_remote = fname
        fname = op.join(tdir, "fig.gif")
        fs.get(fname_remote, fname)

    display.display(display.Image(fname))


def show_anatomical_slices(img_data, title):
    """
    display anatomical slices from midpoint

    based on:
    https://nipy.org/nibabel/coordinate_systems.html
    """

    axial_slice = img_data[:, :, int(img_data.shape[2] / 2)]
    coronal_slice = img_data[:, int(img_data.shape[1] / 2), :]
    sagittal_slice = img_data[int(img_data.shape[0] / 2), :, :]

    fig = plt.figure(constrained_layout=False)
    gs = fig.add_gridspec(nrows=3, ncols=2, wspace=0.01, hspace=0.01)
    ax1 = fig.add_subplot(gs[:-1, :])
    ax1.imshow(axial_slice.T, cmap="gray", origin="lower")
    ax1.axis('off')
    ax2 = fig.add_subplot(gs[2, 0])
    ax2.imshow(coronal_slice.T, cmap="gray", origin="lower")
    ax2.axis('off')
    ax3 = fig.add_subplot(gs[2, 1])
    ax3.imshow(sagittal_slice.T, cmap="gray", origin="lower")
    ax3.axis('off')

    plt.suptitle(title)
    plt.show()



import numpy as np
import pimms
from time import time
import logging

import dipy.tracking.streamline as dts

import AFQ.recognition.utils as abu


logger = logging.getLogger('AFQ')


@pimms.calc("tol", "dist_to_atlas", "vox_dim")
def tolerance_mm_to_vox(img, dist_to_waypoint, input_dist_to_atlas):
    # We need to calculate the size of a voxel, so we can transform
    # from mm to voxel units:
    R = img.affine[0:3, 0:3]
    vox_dim = np.mean(np.diag(np.linalg.cholesky(R.T.dot(R))))

    # Tolerance is set to the square of the distance to the corner
    # because we are using the squared Euclidean distance in calls to
    # `cdist` to make those calls faster.
    if dist_to_waypoint is None:
        tol = dts.dist_to_corner(img.affine)
    else:
        tol = dist_to_waypoint / vox_dim
    dist_to_atlas = int(input_dist_to_atlas / vox_dim)
    return tol, dist_to_atlas, vox_dim


@pimms.calc("fgarray")
def fgarray(tg):
    """
    Streamlines resampled to 20 points.
    """
    logger.info("Resampling Streamlines...")
    start_time = time()
    fg_array = np.array(abu.resample_tg(tg, 20))
    logger.info((
        "Streamlines Resampled "
        f"(time: {time()-start_time}s)"))
    return fg_array


@pimms.calc("crosses")
def crosses(fgarray, img):
    """
    Classify the streamlines by whether they cross the midline.
    Creates a crosses attribute which is an array of booleans. Each boolean
    corresponds to a streamline, and is whether or not that streamline
    crosses the midline.
    """
    # What is the x,y,z coordinate of 0,0,0 in the template space?
    zero_coord = np.dot(np.linalg.inv(img.affine),
                        np.array([0, 0, 0, 1]))

    return np.logical_and(
        np.any(fgarray[:, :, 0] > zero_coord[0], axis=1),
        np.any(fgarray[:, :, 0] < zero_coord[0], axis=1))


# Things that can be calculated for multiple bundles at once
# (i.e., for a whole tractogram) go here
def get_preproc_plan(img, tg, dist_to_waypoint, dist_to_atlas):
    preproc_plan = pimms.Plan(
        tolerance_mm_to_vox=tolerance_mm_to_vox,
        fgarray=fgarray,
        crosses=crosses)
    return preproc_plan(
        img=img, tg=tg,
        dist_to_waypoint=dist_to_waypoint,
        input_dist_to_atlas=dist_to_atlas)



"""
==========================================
Using cloudknot to run pyAFQ on AWS batch:
==========================================
One of the purposes of ``pyAFQ`` is to analyze large-scale openly-available
datasets, such as those in the
`Human Connectome Project <https://www.humanconnectome.org/>`_.

To analyze these datasets, large amounts of compute are needed.
One way to gain access to massive computational power is by using
cloud computing. Here, we will demonstrate
how to use ``pyAFQ`` in the Amazon Web Services cloud.

We will rely on the `AWS Batch Service <https://aws.amazon.com/batch/>`_ ,
and we will submit work into AWS Batch using software that our group
developed called `Cloudknot <https://nrdg.github.io/cloudknot/>`_.
"""

##########################################################################
# Import cloudknot and set the AWS region within which computations will take
# place. Setting a region is important, because if the data that you are
# analyzing is stored in `AWS S3 <https://aws.amazon.com/s3/>`_ in a
# particular region, it is best to run the computation in that region as well.
# That is because AWS charges for inter-region transfer of data.
import cloudknot as ck
ck.set_region('us-east-1')

##########################################################################
# Define the function to use
# --------------------------
# ``Cloudknot`` uses the single program multiple data paradigm of computing.
# This means that the same function will be run on multiple different inputs.
# For example, a ``pyAFQ`` processing function run
# on multiple different subjects in a dataset.
# Below, we define the function that we will use. Notice that
# ``Cloudknot`` functions include the import statements of the dependencies
# used. This is necessary so that ``Cloudknot`` knows
# what dependencies to install into AWS Batch to run this function.


def afq_process_subject(subject):
    # define a function that each job will run
    # In this case, each process does a single subject
    import s3fs
    # all imports must be at the top of the function
    # cloudknot installs the appropriate packages from pip
    from s3bids.utils import S3BIDSStudy
    from AFQ.api.group import GroupAFQ
    import AFQ.definitions.image as afm

    # Download the given subject to your local machine from s3
    # Can find subjects more easily if they are specified in a
    # BIDS participants.tsv file, even if it is sparse
    study_ixi = S3BIDSStudy(
        "my_study",
        "my_study_bucket",
        "my_study_prefix",
        subjects=[subject],
        use_participants_tsv=True,
        anon=False)
    study_ixi.download(
        "local_bids_dir",
        include_derivs=["pipeline_name"])

    # you can optionally provide your own segmentation file
    # in this case, we look for a file with suffix 'seg'
    # in the 'pipeline_name' pipeline,
    # and we consider all non-zero labels to be a part of the brain
    brain_mask_definition = afm.LabelledImageFile(
        suffix='seg', filters={'scope': 'pipeline_name'},
        exclusive_labels=[0])

    # define the api AFQ object
    myafq = GroupAFQ(
        "local_bids_dir",
        preproc_pipeline="pipeline_name",
        brain_mask_definition=brain_mask_definition,
        viz_backend_spec='plotly',  # this will generate both interactive html and GIFs # noqa
        scalars=["dki_fa", "dki_md"])

    # export_all runs the entire pipeline and creates many useful derivates
    myafq.export_all()

    # upload the results to some location on s3
    myafq.upload_to_s3(
        s3fs.S3FileSystem(),
        "my_study_bucket/my_study_prefix/derivatives/afq")


##########################################################################
# Here we provide a list of subjects that we have selected to process
# to randomly select 3 subjects without replacement, instead do:
# subjects = [[1], [2], [3]]
# see the docstring for S3BIDSStudy.__init__ for more information
subjects = ["123456", "123457", "123458"]

##########################################################################
# Defining a ``Knot`` instance
# ---------------------------------
# We instantiate a class instance of the :class:`ck.Knot` class.
# This object will be used to run your jobs.
# The object is instantiated with the `'AmazonS3FullAccess'` policy,
# so that it can write the results
# out to S3, into a bucket that you have write permissions on.
# Setting the `bid_percentage` key-word makes AWS Batch use
# `spot EC2 instances <https://aws.amazon.com/ec2/spot/>`_ for the
# computation. This can result in substantial cost-savings, as spot compute
# instances can cost much less than on-demand instances.
# However, not that spot instances can also
# be evicted, so if completing all of the work is very time-sensitive,
# do not set this key-word argument. Using the `image_github_installs`
# key-word argument will install pyAFQ from GitHub.
# You can also specify other forks and branches to install from.
knot = ck.Knot(
    name='afq-process-subject-201009-0',
    func=afq_process_subject,
    base_image='python:3.8',
    image_github_installs="https://github.com/yeatmanlab/pyAFQ.git",
    pars_policies=('AmazonS3FullAccess',),
    bid_percentage=100)

##########################################################################
# Launching the computation
# --------------------------------
# The :meth:`map` method of the :class:`Knot object maps each of the inputs
# provided as a sequence onto the function and executes the function on each
# one of them in parallel.
result_futures = knot.map(subjects)

##########################################################################
# Once computations have started, you can call the following
# function to view the progress of jobs::
#
#     knot.view_jobs()
#
# You can also view the status of a specific job::
#
#     knot.jobs[0].status


##########################################################################
# When all jobs are finished, remember to use the :meth:`clobber` method to
# destroy all of the AWS resources created by the :class:`Knot`
result_futures.result()
knot.clobber(clobber_pars=True, clobber_repo=True, clobber_image=True)

##########################################################################
# In a second :class:`Knot` object, we use a function that takes the
# resulting profiles of each subject and combines them into one csv file.


def afq_combine_profiles(dummy_argument):
    from AFQ.api import download_and_combine_afq_profiles
    download_and_combine_afq_profiles(
        "my_study_bucket", "my_study_prefix")


knot2 = ck.Knot(
    name='afq_combine_subjects-201009-0',
    func=afq_combine_profiles,
    base_image='python:3.8',
    image_github_installs="https://github.com/yeatmanlab/pyAFQ.git",
    pars_policies=('AmazonS3FullAccess',),
    bid_percentage=100)

##########################################################################
# This knot is called with a dummy argument, which is not used within the
# function itself. The `job_type` key-word argument is used to signal to
# ``Cloudknot`` that only one job is submitted rather than the default
# array of jobs.
result_futures2 = knot2.map(["dummy_argument"], job_type="independent")
result_futures2.result()
knot2.clobber(clobber_pars=True, clobber_repo=True, clobber_image=True)



import tempfile
import os
import os.path as op
import shutil
import subprocess
import gc
import random
import concurrent.futures

import toml

import numpy as np
import numpy.testing as npt
import pytest

import pandas as pd
from pandas.testing import assert_series_equal

import nibabel as nib
import tempfile

import dipy.tracking.utils as dtu
import dipy.tracking.streamline as dts
from dipy.data import get_fnames
from dipy.testing.decorators import xvfb_it
from dipy.io.streamline import load_tractogram

import AFQ.api.bundle_dict as abd
from AFQ.api.group import GroupAFQ, ParallelGroupAFQ
from AFQ.api.participant import ParticipantAFQ
import AFQ.data.fetch as afd
import AFQ.utils.streamlines as aus
import AFQ.utils.bin as afb
from AFQ.definitions.mapping import SynMap, AffMap, SlrMap, IdentityMap
from AFQ.definitions.image import (RoiImage, PFTImage, ImageFile,
                                   TemplateImage, LabelledImageFile)


def touch(fname, times=None):
    with open(fname, 'a'):
        os.utime(fname, times)


def get_temp_hardi():
    tmpdir = tempfile.mkdtemp()
    bids_path = op.join(tmpdir, "stanford_hardi")
    afd.organize_stanford_data(path=tmpdir)

    sub_path = op.join(
        bids_path,
        'derivatives',
        'vistasoft',
        'sub-01',
        'ses-01',
        'dwi')

    return tmpdir, bids_path, sub_path


def create_dummy_data(dmriprep_dir, subject, session=None):
    aff = np.eye(4)
    data = np.ones((10, 10, 10, 6))
    bvecs = np.vstack([np.eye(3), np.eye(3)])
    bvecs[0] = 0
    bvals = np.ones(6) * 1000.
    bvals[0] = 0

    if session is None:
        data_dir = subject
    else:
        data_dir = op.join(subject, session)

    np.savetxt(
        op.join(
            dmriprep_dir, data_dir, 'dwi', 'dwi.bval'),
        bvals)
    np.savetxt(
        op.join(
            dmriprep_dir, data_dir, 'dwi', 'dwi.bvec'),
        bvecs)
    nib.save(
        nib.Nifti1Image(data, aff),
        op.join(
            dmriprep_dir, data_dir, 'dwi', 'dwi.nii.gz'))
    nib.save(
        nib.Nifti1Image(data, aff),
        op.join(
            dmriprep_dir, data_dir, 'anat', 'T1w.nii.gz'))
    nib.save(
        nib.Nifti1Image(data, aff),
        op.join(
            dmriprep_dir, data_dir, 'anat', 'seg.nii.gz'))


def create_dummy_bids_path(n_subjects, n_sessions, share_sessions=True):
    subjects = ['sub-0%s' % (d + 1) for d in range(n_subjects)]

    # Case where there are individual session folders within each subject's
    # folder:
    if n_sessions > 1:
        # create data for n_sessions for each subject
        if share_sessions:
            sessions = ['ses-0%s' % (d + 1) for d in range(n_sessions)]

            bids_dir = tempfile.mkdtemp()

            afd.to_bids_description(
                bids_dir,
                **{"Name": "Dummy",
                   "Subjects": subjects,
                   "Sessions": sessions})

            dmriprep_dir = op.join(bids_dir, "derivatives", "dmriprep")
            os.makedirs(dmriprep_dir)
            afd.to_bids_description(
                dmriprep_dir,
                **{"Name": "Dummy",
                   "PipelineDescription": {"Name": "synthetic"},
                   "GeneratedBy": [{"Name": "synthetic"}]})

            for subject in subjects:
                for session in sessions:
                    for modality in ['anat', 'dwi']:
                        os.makedirs(
                            op.join(dmriprep_dir, subject, session, modality))
                    # Make some dummy data:
                    create_dummy_data(dmriprep_dir, subject, session)
        else:
            # create different sessions for each subject
            sessions = ['ses-0%s' % (d + 1) for d in range(n_subjects)]

            bids_dir = tempfile.mkdtemp()

            afd.to_bids_description(
                bids_dir,
                **{"Name": "Dummy",
                   "Subjects": subjects,
                   "Sessions": sessions})

            dmriprep_dir = op.join(bids_dir, "derivatives", "dmriprep")
            os.makedirs(dmriprep_dir)
            afd.to_bids_description(
                dmriprep_dir,
                **{"Name": "Dummy",
                   "PipelineDescription": {"Name": "synthetic"},
                   "GeneratedBy": [{"Name": "synthetic"}]})

            for d in range(n_subjects):
                subject = subjects[d]
                session = sessions[d]
                for modality in ['anat', 'dwi']:
                    os.makedirs(
                        op.join(dmriprep_dir, subject, session, modality))
                # Make some dummy data:
                create_dummy_data(dmriprep_dir, subject, session)
    else:
        # Don't create session folders at all:
        bids_dir = tempfile.mkdtemp()

        afd.to_bids_description(
            bids_dir,
            **{"Name": "Dummy", "Subjects": subjects})

        dmriprep_dir = op.join(bids_dir, "derivatives", "dmriprep")
        os.makedirs(dmriprep_dir)
        afd.to_bids_description(
            dmriprep_dir,
            **{"Name": "Dummy",
               "PipelineDescription": {"Name": "synthetic"},
               "GeneratedBy": [{"Name": "synthetic"}]})

        for subject in subjects:
            for modality in ['anat', 'dwi']:
                os.makedirs(op.join(dmriprep_dir, subject, modality))
            # Make some dummy data:
            create_dummy_data(dmriprep_dir, subject)

    return bids_dir


def test_AFQ_missing_files():
    tmpdir = tempfile.TemporaryDirectory()
    bids_path = tmpdir.name

    with pytest.raises(
            ValueError,
            match="There must be a dataset_description.json in bids_path"):
        GroupAFQ(bids_path)
    afd.to_bids_description(
        bids_path,
        **{"Name": "Missing", "Subjects": ["sub-01"]})

    with pytest.raises(
            ValueError,
            match=f"No non-json files recognized by pyBIDS in {bids_path}"):
        GroupAFQ(bids_path)

    subses_folder = op.join(
        bids_path,
        "derivatives",
        "otherDeriv",
        'sub-01',
        'ses-01')
    os.makedirs(subses_folder, exist_ok=True)
    afd.to_bids_description(
        op.join(
            bids_path,
            "derivatives",
            "otherDeriv"),
        **{
            "Name": "Missing",
            "PipelineDescription": {"Name": "otherDeriv"},
            "GeneratedBy": [{"Name": "otherDeriv"}]})
    touch(op.join(subses_folder, "sub-01_ses-01_dwi.nii.gz"))

    with pytest.raises(
            ValueError,
            match="No non-json files recognized by pyBIDS"
            + " in the pipeline: missingPipe"):
        GroupAFQ(bids_path, preproc_pipeline="missingPipe")

    os.mkdir(op.join(bids_path, "missingPipe"))
    afd.to_bids_description(
        op.join(bids_path, "missingPipe"), **{
            "Name": "Missing",
            "PipelineDescription": {"Name": "missingPipe"},
            "GeneratedBy": [{"Name": "missingPipe"}]})
    with pytest.raises(
            ValueError,
            match="No non-json files recognized by pyBIDS"
            + " in the pipeline: missingPipe"):
        GroupAFQ(bids_path, preproc_pipeline="missingPipe")


@pytest.mark.nightly_custom
def test_AFQ_custom_tract():
    """
    Test whether AFQ can use tractography from
    import_tract
    """
    _, bids_path, sub_path = get_temp_hardi()
    afd.fetch_stanford_hardi_tractography()

    bundle_info = abd.default18_bd()[
        "Left Superior Longitudinal",
        "Right Superior Longitudinal",
        "Left Arcuate",
        "Right Arcuate",
        "Left Corticospinal",
        "Right Corticospinal",
        "Forceps Minor"]

    # move subsampled tractography into bids folder
    os.rename(
        op.join(
            op.expanduser('~'),
            'AFQ_data',
            'stanford_hardi_tractography',
            'tractography_subsampled.trk'),
        op.join(
            sub_path,
            'subsampled_tractography.trk'
            )
    )
    myafq = GroupAFQ(
        bids_path,
        preproc_pipeline='vistasoft',
        bundle_info=bundle_info,
        import_tract={
            "suffix": "tractography",
            "scope": "vistasoft"
        })

    # equivalent ParticipantAFQ version of call may be useful as reference
    # myafq = ParticipantAFQ(
    #     sub_path + "/sub-01_ses-01_dwi.nii.gz",
    #     sub_path + "/sub-01_ses-01_dwi.bval",
    #     sub_path + "/sub-01_ses-01_dwi.bvec",
    #     output_dir=sub_path,
    #     bundle_info=bundle_names,
    #     import_tract=sub_path + '/subsampled_tractography.trk')

    myafq.export("streamlines")


@pytest.mark.nightly_basic
def test_AFQ_no_derivs():
    """
    Test the initialization of the GroupAFQ object
    """
    bids_path = create_dummy_bids_path(1, 1)
    os.remove(op.join(
        bids_path, "derivatives", "dmriprep", "dataset_description.json"))
    with pytest.raises(
            ValueError,
            match=f"No non-json files recognized by pyBIDS in {bids_path}"):
        GroupAFQ(
            bids_path,
            preproc_pipeline="synthetic")


@pytest.mark.nightly_custom
@xvfb_it
def test_AFQ_fury():
    tmpdir = tempfile.TemporaryDirectory()
    bids_path = op.join(tmpdir.name, "stanford_hardi")
    afd.organize_stanford_data(path=tmpdir.name)

    myafq = GroupAFQ(
        bids_path=bids_path,
        preproc_pipeline='vistasoft',
        viz_backend_spec="fury")
    myafq.export("all_bundles_figure")


@pytest.mark.nightly_pft
def test_AFQ_trx():
    tmpdir = tempfile.TemporaryDirectory()
    bids_path = op.join(tmpdir.name, "stanford_hardi")
    afd.organize_stanford_data(path=tmpdir.name)

    myafq = GroupAFQ(
        bids_path=bids_path,
        preproc_pipeline='vistasoft',
        # should throw warning but not error
        scalars=["dti_fa", "dti_md", ImageFile(suffix="DNE")],
        tracking_params={"trx": True})
    myafq.export("all_bundles_figure")


def test_AFQ_init():
    """
    Test the initialization of the GroupAFQ object
    """
    for n_sessions in [1, 2, 3]:
        if n_sessions == 3:
            # we only need to test all of these once
            participant_labels_to_test = [None, ["01"], ["04"]]
        else:
            participant_labels_to_test = [None]
        for participant_labels in participant_labels_to_test:
            if participant_labels is None:
                n_subjects = 3
            else:
                n_subjects = 1
            bids_path = create_dummy_bids_path(
                n_subjects, n_sessions,
                (n_subjects != n_sessions))

            if participant_labels is not None and\
                    participant_labels[0] == "04":
                with pytest.raises(
                    ValueError,
                    match="No subjects specified in `participant_labels` "
                    + " found in BIDS derivatives folders."
                        + " See above warnings."):
                    myafq = GroupAFQ(
                        bids_path,
                        preproc_pipeline="synthetic",
                        participant_labels=participant_labels)
            else:
                myafq = GroupAFQ(
                    bids_path,
                    preproc_pipeline="synthetic",
                    participant_labels=participant_labels)
                myafq.export("dwi")


@pytest.mark.nightly_basic
def test_AFQ_data():
    """
    Test if API can run without prealign and with only pre-align
    """
    _, bids_path, _ = get_temp_hardi()

    for mapping in [SynMap(use_prealign=False), AffMap()]:
        myafq = GroupAFQ(
            bids_path=bids_path,
            preproc_pipeline='vistasoft',
            mapping_definition=mapping)
        npt.assert_equal(nib.load(myafq.export("b0")["01"]).shape,
                         myafq.export("dwi")["01"].shape[:3])
        npt.assert_equal(nib.load(myafq.export("b0")["01"]).shape,
                         nib.load(myafq.export("dti_params")["01"]).shape[:3])
        myafq.export("rois")
        myafq.cmd_outputs()
        assert len(os.listdir(op.join(
            myafq.afq_path, "sub-01", "ses-01"))) == 0


@pytest.mark.nightly_anisotropic
def test_AFQ_anisotropic():
    """
    Test if API can run using anisotropic power map registration
    with a specific selection of b vals
    """
    _, bids_path, _ = get_temp_hardi()
    myafq = GroupAFQ(
        bids_path=bids_path,
        preproc_pipeline='vistasoft',
        min_bval=1990,
        max_bval=2010,
        b0_threshold=50,
        reg_template_spec="mni_T1",
        reg_subject_spec="power_map")

    gtab = myafq.export("gtab")["01"]

    # check the b0s mask is correct
    b0s_mask = np.zeros(160, dtype=bool)
    b0s_mask[0:10] = True
    npt.assert_equal(gtab.b0s_mask, b0s_mask)

    # check that only b values in the b val range passed
    bvals_in_range = \
        np.logical_and((gtab.bvals > 1990), (gtab.bvals < 2010))
    bvals_in_range_or_0 = \
        np.logical_or(bvals_in_range, gtab.b0s_mask)
    npt.assert_equal(bvals_in_range_or_0, np.ones(160, dtype=bool))

    # check that the apm map was made
    myafq.export("mapping")
    assert op.exists(op.join(
        myafq.export("output_dir")["01"],
        'sub-01_ses-01_odfmodel-CSD_desc-APM_dwi.nii.gz'))


def test_API_type_checking():
    _, bids_path, _ = get_temp_hardi()
    seed = 2022
    np.random.seed(seed)
    # Note that the ordering of these "with pytest.raises"
    # statements is important. Later tests will use
    # the sucessful results of previous ones. IE,
    # do not put a test of brain masks in front of a test
    # of all_bundles_figure, because it will use the brain mask
    # generated by the all_bundles_figure test isntead of testing
    # a new brain mask.
    with pytest.raises(
            TypeError,
            match="bids_path must be a string"):
        GroupAFQ(2)

    with pytest.raises(
            TypeError,
            match=(
                "import_tract must be"
                " either a dict or a str")):
        myafq = GroupAFQ(
            bids_path,
            preproc_pipeline='vistasoft',
            import_tract=["dwi"])
        myafq.export("streamlines")
    del myafq

    with pytest.raises(
            TypeError,
            match="brain_mask_definition must be a Definition"):
        myafq = GroupAFQ(
            bids_path,
            preproc_pipeline='vistasoft',
            brain_mask_definition="not a brain mask")

    with pytest.raises(
            ValueError,
            match=r"No file found with these parameters:\n*"):
        myafq = GroupAFQ(
            bids_path,
            preproc_pipeline='vistasoft',
            brain_mask_definition=ImageFile(
                suffix='dne_dne',
                filters={'scope': 'dne_dne'}))
        myafq.export("brain_mask")

    with pytest.raises(
            TypeError,
            match=(
                "bundle_info must be"
                " a dict, or a BundleDict")):
        myafq = GroupAFQ(
            bids_path,
            preproc_pipeline='vistasoft',
            bundle_info=[2, 3])
        myafq.export("bundle_dict")
    del myafq

    with pytest.raises(
            ValueError,
            match=(
                "Fatal: No bundles recognized.")):
        myafq = GroupAFQ(
            bids_path,
            preproc_pipeline='vistasoft',
            mapping_definition=IdentityMap(),
            tracking_params={
                "n_seeds": 10,
                "rng_seed": seed,
                "random_seeds": True,
                "directions": "det",
                "odf_model": "DTI"},
            bundle_info=abd.default18_bd()["Left Arcuate", "Right Arcuate"])
        myafq.export("bundles")
    del myafq

    with pytest.raises(
            TypeError,
            match="viz_backend_spec must contain either 'fury' or 'plotly'"):
        myafq = GroupAFQ(
            bids_path,
            preproc_pipeline='vistasoft',
            viz_backend_spec="matplotlib")
        myafq.export("viz_backend")
    del myafq


def test_AFQ_slr():
    """
    Test if API can run using slr map
    """
    seed = 2022
    random.seed(seed)

    afd.read_stanford_hardi_tractography()

    _, bids_path, _ = get_temp_hardi()
    bd = abd.default18_bd()[(
        "Left Corticospinal",
        "Right Corticospinal")]

    myafq = GroupAFQ(
        bids_path=bids_path,
        preproc_pipeline='vistasoft',
        reg_subject_spec='subject_sls',
        reg_template_spec='hcp_atlas',
        import_tract=op.join(
            op.expanduser('~'),
            'AFQ_data',
            'stanford_hardi_tractography',
            'full_segmented_cleaned_tractography.trk'),
        segmentation_params={
            "dist_to_waypoint": 10,
            "parallel_segmentation": {"engine": "serial"}},
        bundle_info=bd,
        mapping_definition=SlrMap(slr_kwargs={
            "rng": np.random.RandomState(seed)}))

    seg_sft = aus.SegmentedSFT.fromfile(
        myafq.export("bundles")["01"])
    npt.assert_(len(seg_sft.get_bundle(
        'Left Corticospinal').streamlines) > 0)


@pytest.mark.nightly_reco
def test_AFQ_reco():
    """
    Test if API can run segmentation with recobundles
    """
    _, bids_path, _ = get_temp_hardi()
    myafq = GroupAFQ(
        bids_path=bids_path,
        preproc_pipeline='vistasoft',
        viz_backend_spec="plotly",
        profile_weights="median",
        bundle_info=abd.reco_bd(16),
        segmentation_params={
            'rng': 42})

    seg_sft = aus.SegmentedSFT.fromfile(
        myafq.export("bundles")["01"])
    npt.assert_(len(seg_sft.get_bundle('CCMid').streamlines) > 0)
    myafq.export_all()


@pytest.mark.nightly_reco80
def test_AFQ_reco80():
    """
    Test API segmentation with the 80-bundle atlas
    """
    _, bids_path, _ = get_temp_hardi()

    tracking_params = dict(
        odf_model="csd",
        n_seeds=10000,
        random_seeds=True,
        rng_seed=42)

    myafq = GroupAFQ(
        bids_path=bids_path,
        preproc_pipeline='vistasoft',
        tracking_params=tracking_params,
        bundle_info=abd.reco_bd(16),
        segmentation_params={
            'rng': 42})

    seg_sft = aus.SegmentedSFT.fromfile(
        myafq.export("bundles")["01"])
    npt.assert_(len(seg_sft.get_bundle('CCMid').streamlines) > 0)


def test_AFQ_pydra():
    _, bids_path = afd.fetch_hbn_preproc(["NDARAA948VFH", "NDARAV554TP2"])
    pga = ParallelGroupAFQ(bids_path, preproc_pipeline="qsiprep")
    pga.export("dti_fa")


def test_AFQ_filterb():
    _, bids_path, _ = get_temp_hardi()
    myafq = GroupAFQ(
        bids_path=bids_path,
        preproc_pipeline='vistasoft',
        max_bval=1000)
    myafq.export("b0")


@pytest.mark.nightly_pft
def test_AFQ_pft():
    """
    Test pft interface for AFQ
    """
    _, bids_path, sub_path = get_temp_hardi()

    bundle_names = abd.default18_bd()[
        "Left Superior Longitudinal",
        "Right Superior Longitudinal",
        "Left Arcuate",
        "Right Arcuate",
        "Left Corticospinal",
        "Right Corticospinal",
        "Forceps Minor"]

    f_pve_csf, f_pve_gm, f_pve_wm = get_fnames('stanford_pve_maps')
    os.rename(f_pve_wm, op.join(sub_path,
                                "sub-01_ses-01_label-WM_probseg.nii.gz"))
    os.rename(f_pve_gm, op.join(sub_path,
                                "sub-01_ses-01_label-GM_probseg.nii.gz"))
    os.rename(f_pve_csf, op.join(sub_path,
                                 "sub-01_ses-01_label-CSF_probseg.nii.gz"))

    stop_mask = PFTImage(
        ImageFile(suffix="probseg", filters={"label": "WM"}),
        ImageFile(suffix="probseg", filters={"label": "GM"}),
        ImageFile(suffix="probseg", filters={"label": "CSF"}))
    t_output_dir = tempfile.TemporaryDirectory()

    myafq = GroupAFQ(
        bids_path,
        preproc_pipeline='vistasoft',
        bundle_info=bundle_names,
        output_dir=t_output_dir.name,
        tracking_params={
            "stop_mask": stop_mask,
            "stop_threshold": "CMC",
            "tracker": "pft",
            "maxlen": 150,
        })
    sl_file = myafq.export("streamlines")["01"]
    dwi_file = myafq.export("dwi")["01"]
    sls = load_tractogram(
        sl_file,
        dwi_file,
        bbox_valid_check=False,
        trk_header_check=False).streamlines
    for sl in sls:
        # double the maxlen, due to step size of 0.5
        assert len(sl) <= 300


@pytest.mark.nightly_custom
def test_AFQ_custom_subject_reg():
    """
    Test custom subject registration using GroupAFQ object
    """
    # make first temproary directory to generate b0
    _, bids_path, sub_path = get_temp_hardi()

    bundle_info = abd.default18_bd()[
        "Left Superior Longitudinal",
        "Right Superior Longitudinal",
        "Left Arcuate",
        "Right Arcuate",
        "Left Corticospinal",
        "Right Corticospinal",
        "Forceps Minor"]

    b0_file = GroupAFQ(
        bids_path,
        preproc_pipeline='vistasoft',
        bundle_info=bundle_info).export("b0")["01"]

    # make a different temporary directly to test this custom file in
    _, bids_path, sub_path = get_temp_hardi()

    os.rename(b0_file, op.join(sub_path, "sub-01_ses-01_customb0.nii.gz"))

    myafq = GroupAFQ(
        bids_path,
        preproc_pipeline='vistasoft',
        bundle_info=bundle_info,
        reg_template_spec="mni_T2",
        reg_subject_spec=ImageFile(
            suffix="customb0",
            filters={"scope": "vistasoft"}))
    myafq.export("rois")


# Requires large download
@pytest.mark.nightly
def test_AFQ_FA():
    """
    Test if API can run registeration with FA
    """
    _, bids_path, _ = get_temp_hardi()
    myafq = GroupAFQ(
        bids_path=bids_path,
        preproc_pipeline='vistasoft',
        reg_template_spec='dti_fa_template',
        reg_subject_spec='dti_fa_subject')
    myafq.export("rois")


@pytest.mark.nightly
def test_multib_profile():
    """
    Test using API to profile DKI/fwDTI
    """
    tmpdir = tempfile.TemporaryDirectory()
    afd.organize_cfin_data(path=tmpdir.name)
    myafq = GroupAFQ(bids_path=op.join(tmpdir.name, 'cfin_multib'),
                    preproc_pipeline='dipy')
    myafq.export("dki_fa")
    myafq.export("dki_md")
    myafq.export("fwdti_fa")
    myafq.export("fwdti_md")


def test_auto_cli():
    tmpdir = tempfile.TemporaryDirectory()
    config_file = op.join(tmpdir.name, 'test.toml')

    arg_dict = afb.func_dict_to_arg_dict()
    arg_dict['BIDS_PARAMS']['bids_path']['default'] = tmpdir.name
    afb.generate_config(config_file, arg_dict, False)
    with pytest.raises(
            ValueError,
            match="There must be a dataset_description.json in bids_path"):
        afb.parse_config_run_afq(config_file, arg_dict, False)


@pytest.mark.skip(reason="causes segmentation fault")
def test_run_using_auto_cli():
    tmpdir, bids_path, _ = get_temp_hardi()
    config_file = op.join(tmpdir.name, 'test.toml')

    arg_dict = afb.func_dict_to_arg_dict()

    # set our custom defaults for the toml file
    # It is easier to edit them here, than to parse the file and edit them
    # after the file is written
    arg_dict['BIDS_PARAMS']['bids_path']['default'] = bids_path
    arg_dict['BIDS_PARAMS']['dmriprep']['default'] = 'vistasoft'
    arg_dict['DATA']['bundle_info']['default'] = abd.default18_bd()[(
        "Left Corticospinal")]
    arg_dict['TRACTOGRAPHY_PARAMS']['n_seeds']['default'] = 500
    arg_dict['TRACTOGRAPHY_PARAMS']['random_seeds']['default'] = True

    afb.generate_config(config_file, arg_dict, False)
    afb.parse_config_run_afq(config_file, arg_dict, False)


def test_AFQ_data_waypoint():
    """
    Test with some actual data again, this time for track segmentation
    """
    tmpdir, bids_path, _ = get_temp_hardi()
    t1_path = op.join(tmpdir, "T1.nii.gz")
    t1_path_other = op.join(tmpdir, "T1-untransformed.nii.gz")
    nib.save(
        afd.read_mni_template(mask=True, weight="T1w"),
        t1_path)
    shutil.copy(t1_path, t1_path_other)

    vista_folder = op.join(
        bids_path,
        "derivatives/vistasoft/sub-01/ses-01/dwi")

    # Prepare LV1 ROI
    lv1_files, lv1_folder = afd.fetch_stanford_hardi_lv1()
    lv1_fname = op.join(
        lv1_folder,
        list(lv1_files.keys())[0])
    bundle_names = [
        "Left Superior Longitudinal",
        "Right Superior Longitudinal",
        "Left Arcuate",
        "Right Arcuate",
        "Left Corticospinal",
        "Right Corticospinal",
        "Forceps Minor"]
    bundle_info = abd.default18_bd()[bundle_names]

    bundle_info.resample_subject_to = nib.load(
            op.join(vista_folder, "sub-01_ses-01_dwi.nii.gz"))

    # Test when we have endpoint ROIs only
    afq_templates = afd.read_templates(as_img=False)
    bundle_info["Left Superior Longitudinal"] = {
        "start": afq_templates["SLF_L_start"],
        "end": afq_templates["SLF_L_end"],
        "prob_map": afq_templates["SLF_L_prob_map"]
    }

    bundle_info["LV1"] = {
        "start": lv1_fname,
        "space": "subject"
    }

    tracking_params = dict(odf_model="csd",
                           seed_mask=RoiImage(),
                           n_seeds=200,
                           random_seeds=True,
                           rng_seed=42)
    segmentation_params = dict(return_idx=True)

    afq_folder = op.join(bids_path, "derivatives/afq/sub-01/ses-01")
    os.makedirs(afq_folder, exist_ok=True)
    myafq = ParticipantAFQ(
        op.join(vista_folder, "sub-01_ses-01_dwi.nii.gz"),
        op.join(vista_folder, "sub-01_ses-01_dwi.bval"),
        op.join(vista_folder, "sub-01_ses-01_dwi.bvec"),
        afq_folder,
        bundle_info=bundle_info,
        scalars=[
            "dti_FA",
            "dti_MD",
            "dti_GA",
            "dti_lt2",
            ImageFile(path=t1_path_other),
            TemplateImage(t1_path)],
        n_points_profile=50,
        tracking_params=tracking_params,
        segmentation_params=segmentation_params)

    # Replace the mapping and streamlines with precomputed:
    file_dict = afd.read_stanford_hardi_tractography()
    mapping = file_dict['mapping.nii.gz']
    streamlines = file_dict['tractography_subsampled.trk']
    dwi_affine = myafq.export("dwi_affine")
    streamlines = dts.Streamlines(
        dtu.transform_tracking_output(
            [s for s in streamlines if s.shape[0] > 100],
            np.linalg.inv(dwi_affine)))

    mapping_file = op.join(
        myafq.export("output_dir"),
        'sub-01_ses-01_desc-mapping_from-DWI_to-MNI_xform.nii.gz')
    nib.save(mapping, mapping_file)
    reg_prealign_file = op.join(
        myafq.export("output_dir"),
        'sub-01_ses-01_desc-prealign_from-DWI_to-MNI_xform.npy')
    np.save(reg_prealign_file, np.eye(4))

    # Test ROI exporting:
    myafq.export("rois")
    assert op.exists(op.join(
        myafq.export("output_dir"),
        'ROIs',
        'sub-01_ses-01_space-subject_desc-RightCorticospinalinclude1_mask.json'))  # noqa

    seg_sft = aus.SegmentedSFT.fromfile(
        myafq.export("bundles"))
    npt.assert_(len(seg_sft.get_bundle(
        'Left Corticospinal').streamlines) > 0)

    # Test bundles exporting:
    myafq.export("indiv_bundles")
    assert op.exists(op.join(
        myafq.export("output_dir"),
        'bundles',
        'sub-01_ses-01_coordsys-RASMM_trkmethod-probCSD_recogmethod-AFQ_desc-LeftCorticospinal_tractography.trk'))  # noqa

    tract_profile_fname = myafq.export("profiles")
    tract_profiles = pd.read_csv(tract_profile_fname)

    assert tract_profiles.select_dtypes(include=[np.number]).sum().sum() != 0
    assert tract_profiles.shape[0] >= 200
    assert tract_profiles.shape[1] == 9

    myafq.export("indiv_bundles_figures")
    assert op.exists(op.join(
        myafq.export("output_dir"),
        "viz_bundles",
        'sub-01_ses-01_coordsys-RASMM_trkmethod-probCSD_recogmethod-AFQ_desc-LeftSuperiorLongitudinalviz_dwi.html'))  # noqa

    assert op.exists(op.join(
        myafq.export("output_dir"),
        "viz_bundles",
        'sub-01_ses-01_coordsys-RASMM_trkmethod-probCSD_recogmethod-AFQ_desc-LeftSuperiorLongitudinalviz_dwi.html'))  # noqa

    # Before we run the CLI, we'll remove the bundles and ROI folders, to see
    # that the CLI generates them
    shutil.rmtree(op.join(myafq.export("output_dir"),
                          'bundles'))

    shutil.rmtree(op.join(myafq.export("output_dir"),
                          'ROIs'))
    os.remove(tract_profile_fname)

    # save memory
    output_dir = myafq.export("output_dir")
    del myafq
    gc.collect()

    # Test the CLI:
    print("Running the CLI:")

    # Set up config to use the same parameters as above:
    # ROI mask needs to be put in quotes in config
    tracking_params = dict(odf_model="CSD",
                           seed_mask="RoiImage()",
                           n_seeds=200,
                           random_seeds=True,
                           rng_seed=42)
    bundle_dict_as_str = (
        'default18_bd()['
        '"Left Superior Longitudinal",'
        '"Right Superior Longitudinal",'
        '"Left Arcuate",'
        '"Right Arcuate",'
        '"Left Corticospinal",'
        '"Right Corticospinal",'
        '"Forceps Minor"]'
        '+ BundleDict({"LV1": {"start": '
        f'"{lv1_fname}", '
        '"space": "subject"}})')
    config = dict(
        BIDS_PARAMS=dict(
            bids_path=bids_path,
            preproc_pipeline='vistasoft'),
        DATA=dict(
            bundle_info=bundle_dict_as_str),
        SEGMENTATION=dict(
            n_points_profile=50,
            scalars=[
                "dti_fa",
                "dti_md",
                "dti_ga",
                "dti_lt2",
                f"ImageFile('{t1_path_other}')",
                f"TemplateImage('{t1_path}')"]),
        TRACTOGRAPHY_PARAMS=tracking_params,
        SEGMENTATION_PARAMS=segmentation_params)

    config_file = op.join(tmpdir, "afq_config.toml")
    with open(config_file, 'w') as ff:
        toml.dump(config, ff)

    cmd = f"pyAFQ -v {config_file}"
    completed_process = subprocess.run(
        cmd, shell=True, capture_output=True)
    if completed_process.returncode != 0:
        print(completed_process.stdout)
    print(completed_process.stderr)
    assert completed_process.returncode == 0
    # The tract profiles should already exist from the CLI Run:
    from_file = pd.read_csv(tract_profile_fname)

    assert from_file.shape[0] >= 200
    assert from_file.shape[1] == 9
    assert_series_equal(tract_profiles['dti_fa'], from_file['dti_fa'])

    # Make sure the CLI did indeed generate these:
    assert op.exists(op.join(
        output_dir,
        'ROIs',
        'sub-01_ses-01_space-subject_desc-RightSuperiorLongitudinalinclude1_mask.json'))  # noqa

    assert op.exists(op.join(
        output_dir,
        'bundles',
        'sub-01_ses-01_coordsys-RASMM_trkmethod-probCSD_recogmethod-AFQ_desc-RightSuperiorLongitudinal_tractography.trk'))  # noqa
